# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: record.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0crecord.proto\x12\x07Records\"\\\n\nRecordType\x12\x14\n\x0crecordTypeId\x18\x01 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t\x12\'\n\x05scope\x18\x03 \x01(\x0e\x32\x18.Records.RecordTypeScope\"H\n\x12RecordTypesRequest\x12\x10\n\x08standard\x18\x01 \x01(\x08\x12\x0c\n\x04user\x18\x02 \x01(\x08\x12\x12\n\nenterprise\x18\x03 \x01(\x08\"\x88\x01\n\x13RecordTypesResponse\x12(\n\x0brecordTypes\x18\x01 \x03(\x0b\x32\x13.Records.RecordType\x12\x17\n\x0fstandardCounter\x18\x02 \x01(\x05\x12\x13\n\x0buserCounter\x18\x03 \x01(\x05\x12\x19\n\x11\x65nterpriseCounter\x18\x04 \x01(\x05\"A\n\x18RecordTypeModifyResponse\x12\x14\n\x0crecordTypeId\x18\x01 \x01(\x05\x12\x0f\n\x07\x63ounter\x18\x02 \x01(\x05\"=\n\x11RecordsGetRequest\x12\x13\n\x0brecord_uids\x18\x01 \x03(\x0c\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\xd1\x01\n\x06Record\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12/\n\x0frecord_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\r\n\x05\x65xtra\x18\x05 \x01(\x0c\x12\x0f\n\x07version\x18\x06 \x01(\x05\x12\x1c\n\x14\x63lient_modified_time\x18\x07 \x01(\x03\x12\x10\n\x08revision\x18\x08 \x01(\x03\x12\x10\n\x08\x66ile_ids\x18\t \x03(\x0c\"M\n\x0f\x46olderRecordKey\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_uid\x18\x02 \x01(\x0c\x12\x12\n\nrecord_key\x18\x03 \x01(\x0c\"a\n\x06\x46older\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x12\n\nfolder_key\x18\x02 \x01(\x0c\x12/\n\x0f\x66older_key_type\x18\x03 \x01(\x0e\x32\x16.Records.RecordKeyType\"\x95\x01\n\x04Team\x12\x10\n\x08team_uid\x18\x01 \x01(\x0c\x12\x10\n\x08team_key\x18\x02 \x01(\x0c\x12\x18\n\x10team_private_key\x18\x03 \x01(\x0c\x12-\n\rteam_key_type\x18\x04 \x01(\x0e\x32\x16.Records.RecordKeyType\x12 \n\x07\x66olders\x18\x05 \x03(\x0b\x32\x0f.Records.Folder\"\xac\x01\n\x12RecordsGetResponse\x12 \n\x07records\x18\x01 \x03(\x0b\x32\x0f.Records.Record\x12\x34\n\x12\x66older_record_keys\x18\x02 \x03(\x0b\x32\x18.Records.FolderRecordKey\x12 \n\x07\x66olders\x18\x03 \x03(\x0b\x32\x0f.Records.Folder\x12\x1c\n\x05teams\x18\x04 \x03(\x0b\x32\r.Records.Team\"4\n\nRecordLink\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\",\n\x0bRecordAudit\x12\x0f\n\x07version\x18\x01 \x01(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"\xa0\x02\n\tRecordAdd\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12.\n\x0b\x66older_type\x18\x06 \x01(\x0e\x32\x19.Records.RecordFolderType\x12\x12\n\nfolder_uid\x18\x07 \x01(\x0c\x12\x12\n\nfolder_key\x18\x08 \x01(\x0c\x12)\n\x0crecord_links\x18\t \x03(\x0b\x32\x13.Records.RecordLink\x12#\n\x05\x61udit\x18\n \x01(\x0b\x32\x14.Records.RecordAudit\"M\n\x11RecordsAddRequest\x12#\n\x07records\x18\x01 \x03(\x0b\x32\x12.Records.RecordAdd\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\xea\x01\n\x0cRecordUpdate\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x02 \x01(\x03\x12\x10\n\x08revision\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12-\n\x10record_links_add\x18\x06 \x03(\x0b\x32\x13.Records.RecordLink\x12\x1b\n\x13record_links_remove\x18\x07 \x03(\x0c\x12#\n\x05\x61udit\x18\x08 \x01(\x0b\x32\x14.Records.RecordAudit\"S\n\x14RecordsUpdateRequest\x12&\n\x07records\x18\x01 \x03(\x0b\x32\x15.Records.RecordUpdate\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\x8e\x01\n\x17RecordFileForConversion\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x14\n\x0c\x66ile_file_id\x18\x02 \x01(\t\x12\x15\n\rthumb_file_id\x18\x03 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x12\n\nrecord_key\x18\x05 \x01(\x0c\x12\x10\n\x08link_key\x18\x06 \x01(\x0c\"J\n\x19RecordFolderForConversion\x12\x12\n\nfolder_uid\x18\x01 \x01(\x0c\x12\x19\n\x11record_folder_key\x18\x02 \x01(\x0c\"\x92\x02\n\x11RecordConvertToV3\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x02 \x01(\x03\x12\x10\n\x08revision\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x17\n\x0fnon_shared_data\x18\x05 \x01(\x0c\x12#\n\x05\x61udit\x18\x06 \x01(\x0b\x32\x14.Records.RecordAudit\x12\x35\n\x0brecord_file\x18\x07 \x03(\x0b\x32 .Records.RecordFileForConversion\x12\x36\n\nfolder_key\x18\x08 \x03(\x0b\x32\".Records.RecordFolderForConversion\"]\n\x19RecordsConvertToV3Request\x12+\n\x07records\x18\x01 \x03(\x0b\x32\x1a.Records.RecordConvertToV3\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\'\n\x14RecordsRemoveRequest\x12\x0f\n\x07records\x18\x01 \x03(\x0c\">\n\x0cRecordRevert\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x1a\n\x12revert_to_revision\x18\x02 \x01(\x03\">\n\x14RecordsRevertRequest\x12&\n\x07records\x18\x01 \x03(\x0b\x32\x15.Records.RecordRevert\"c\n\x0fRecordLinkError\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12+\n\x06status\x18\x02 \x01(\x0e\x32\x1b.Records.RecordModifyResult\x12\x0f\n\x07message\x18\x03 \x01(\t\"\x95\x01\n\x12RecordModifyStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12+\n\x06status\x18\x02 \x01(\x0e\x32\x1b.Records.RecordModifyResult\x12\x0f\n\x07message\x18\x03 \x01(\t\x12-\n\x0blink_errors\x18\x04 \x03(\x0b\x32\x18.Records.RecordLinkError\"W\n\x15RecordsModifyResponse\x12,\n\x07records\x18\x01 \x03(\x0b\x32\x1b.Records.RecordModifyStatus\x12\x10\n\x08revision\x18\x02 \x01(\x03\"Y\n\x12RecordAddAuditData\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x0f\n\x07version\x18\x04 \x01(\x05\"C\n\x13\x41\x64\x64\x41uditDataRequest\x12,\n\x07records\x18\x01 \x03(\x0b\x32\x1b.Records.RecordAddAuditData\"a\n\x04\x46ile\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\x12\x10\n\x08\x66ileSize\x18\x04 \x01(\x03\x12\x11\n\tthumbSize\x18\x05 \x01(\x05\"D\n\x0f\x46ilesAddRequest\x12\x1c\n\x05\x66iles\x18\x01 \x03(\x0b\x32\r.Records.File\x12\x13\n\x0b\x63lient_time\x18\x02 \x01(\x03\"\xa7\x01\n\rFileAddStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12&\n\x06status\x18\x02 \x01(\x0e\x32\x16.Records.FileAddResult\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x12\n\nparameters\x18\x04 \x01(\t\x12\x1c\n\x14thumbnail_parameters\x18\x05 \x01(\t\x12\x1b\n\x13success_status_code\x18\x06 \x01(\x05\"K\n\x10\x46ilesAddResponse\x12%\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x16.Records.FileAddStatus\x12\x10\n\x08revision\x18\x02 \x01(\x03\"f\n\x0f\x46ilesGetRequest\x12\x13\n\x0brecord_uids\x18\x01 \x03(\x0c\x12\x16\n\x0e\x66or_thumbnails\x18\x02 \x01(\x08\x12&\n\x1e\x65mergency_access_account_owner\x18\x03 \x01(\t\"\xa2\x01\n\rFileGetStatus\x12\x12\n\nrecord_uid\x18\x01 \x01(\x0c\x12&\n\x06status\x18\x02 \x01(\x0e\x32\x16.Records.FileGetResult\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x1b\n\x13success_status_code\x18\x04 \x01(\x05\x12+\n\x0b\x66ileKeyType\x18\x05 \x01(\x0e\x32\x16.Records.RecordKeyType\"9\n\x10\x46ilesGetResponse\x12%\n\x05\x66iles\x18\x01 \x03(\x0b\x32\x16.Records.FileGetStatus\"h\n\x15\x41pplicationAddRequest\x12\x0f\n\x07\x61pp_uid\x18\x01 \x01(\x0c\x12\x12\n\nrecord_key\x18\x02 \x01(\x0c\x12\x1c\n\x14\x63lient_modified_time\x18\x03 \x01(\x03\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\"\x88\x01\n\"GetRecordDataWithAccessInfoRequest\x12\x12\n\nclientTime\x18\x01 \x01(\x03\x12\x11\n\trecordUid\x18\x02 \x03(\x0c\x12;\n\x14recordDetailsInclude\x18\x03 \x01(\x0e\x32\x1d.Records.RecordDetailsInclude\"\xab\x01\n\x0eUserPermission\x12\x10\n\x08username\x18\x01 \x01(\t\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x12\n\nshareAdmin\x18\x03 \x01(\x08\x12\x10\n\x08sharable\x18\x04 \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x05 \x01(\x08\x12\x18\n\x10\x61waitingApproval\x18\x06 \x01(\x08\x12\x12\n\nexpiration\x18\x07 \x01(\x03\x12\x12\n\naccountUid\x18\x08 \x01(\x0c\"}\n\x16SharedFolderPermission\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\x12\x12\n\nresharable\x18\x02 \x01(\x08\x12\x10\n\x08\x65\x64itable\x18\x03 \x01(\x08\x12\x10\n\x08revision\x18\x04 \x01(\x03\x12\x12\n\nexpiration\x18\x05 \x01(\x03\"\xe8\x02\n\nRecordData\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12\x0f\n\x07version\x18\x02 \x01(\x05\x12\x0e\n\x06shared\x18\x03 \x01(\x08\x12\x1b\n\x13\x65ncryptedRecordData\x18\x04 \x01(\t\x12\x1a\n\x12\x65ncryptedExtraData\x18\x05 \x01(\t\x12\x1a\n\x12\x63lientModifiedTime\x18\x06 \x01(\x03\x12\x15\n\rnonSharedData\x18\x07 \x01(\t\x12-\n\x10linkedRecordData\x18\x08 \x03(\x0b\x32\x13.Records.RecordData\x12\x0e\n\x06\x66ileId\x18\t \x03(\x0c\x12\x10\n\x08\x66ileSize\x18\n \x01(\x03\x12\x15\n\rthumbnailSize\x18\x0b \x01(\x03\x12-\n\rrecordKeyType\x18\x0c \x01(\x0e\x32\x16.Records.RecordKeyType\x12\x11\n\trecordKey\x18\r \x01(\x0c\x12\x11\n\trecordUid\x18\x0e \x01(\x0c\"\xc8\x01\n\x18RecordDataWithAccessInfo\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\'\n\nrecordData\x18\x02 \x01(\x0b\x32\x13.Records.RecordData\x12/\n\x0euserPermission\x18\x03 \x03(\x0b\x32\x17.Records.UserPermission\x12?\n\x16sharedFolderPermission\x18\x04 \x03(\x0b\x32\x1f.Records.SharedFolderPermission\"\x89\x01\n#GetRecordDataWithAccessInfoResponse\x12\x43\n\x18recordDataWithAccessInfo\x18\x01 \x03(\x0b\x32!.Records.RecordDataWithAccessInfo\x12\x1d\n\x15noPermissionRecordUid\x18\x02 \x03(\x0c\"j\n\x12IsObjectShareAdmin\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\x12\x0f\n\x07isAdmin\x18\x02 \x01(\x08\x12\x36\n\nobjectType\x18\x03 \x01(\x0e\x32\".Records.CheckShareAdminObjectType\"H\n\rAmIShareAdmin\x12\x37\n\x12isObjectShareAdmin\x18\x01 \x03(\x0b\x32\x1b.Records.IsObjectShareAdmin\"\xbc\x01\n\x18RecordShareUpdateRequest\x12.\n\x0f\x61\x64\x64SharedRecord\x18\x01 \x03(\x0b\x32\x15.Records.SharedRecord\x12\x31\n\x12updateSharedRecord\x18\x02 \x03(\x0b\x32\x15.Records.SharedRecord\x12\x31\n\x12removeSharedRecord\x18\x03 \x03(\x0b\x32\x15.Records.SharedRecord\x12\n\n\x02pt\x18\x04 \x01(\t\"\xd5\x01\n\x0cSharedRecord\x12\x12\n\ntoUsername\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x04 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x05 \x01(\x0c\x12\x10\n\x08\x65\x64itable\x18\x06 \x01(\x08\x12\x11\n\tshareable\x18\x07 \x01(\x08\x12\x10\n\x08transfer\x18\x08 \x01(\x08\x12\x11\n\tuseEccKey\x18\t \x01(\x08\x12\x17\n\x0fremoveVaultData\x18\n \x01(\x08\"\xd5\x01\n\x19RecordShareUpdateResponse\x12:\n\x15\x61\x64\x64SharedRecordStatus\x18\x01 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\x12=\n\x18updateSharedRecordStatus\x18\x02 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\x12=\n\x18removeSharedRecordStatus\x18\x03 \x03(\x0b\x32\x1b.Records.SharedRecordStatus\"Z\n\x12SharedRecordStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\x12\x0f\n\x07message\x18\x03 \x01(\t\x12\x10\n\x08username\x18\x04 \x01(\t\"G\n\x1bGetRecordPermissionsRequest\x12\x12\n\nrecordUids\x18\x01 \x03(\x0c\x12\x14\n\x0cisShareAdmin\x18\x02 \x01(\x08\"T\n\x1cGetRecordPermissionsResponse\x12\x34\n\x11recordPermissions\x18\x01 \x03(\x0b\x32\x19.Records.RecordPermission\"l\n\x10RecordPermission\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\r\n\x05owner\x18\x02 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x03 \x01(\x08\x12\x10\n\x08\x63\x61nShare\x18\x04 \x01(\x08\x12\x13\n\x0b\x63\x61nTransfer\x18\x05 \x01(\x08\"h\n\x16GetShareObjectsRequest\x12\x11\n\tstartWith\x18\x01 \x01(\t\x12\x10\n\x08\x63ontains\x18\x02 \x01(\t\x12\x10\n\x08\x66iltered\x18\x03 \x01(\x08\x12\x17\n\x0fsharedFolderUid\x18\x04 \x01(\x0c\"\xe7\x02\n\x17GetShareObjectsResponse\x12.\n\x12shareRelationships\x18\x01 \x03(\x0b\x32\x12.Records.ShareUser\x12,\n\x10shareFamilyUsers\x18\x02 \x03(\x0b\x32\x12.Records.ShareUser\x12\x30\n\x14shareEnterpriseUsers\x18\x03 \x03(\x0b\x32\x12.Records.ShareUser\x12&\n\nshareTeams\x18\x04 \x03(\x0b\x32\x12.Records.ShareTeam\x12(\n\x0cshareMCTeams\x18\x05 \x03(\x0b\x32\x12.Records.ShareTeam\x12\x32\n\x16shareMCEnterpriseUsers\x18\x06 \x03(\x0b\x32\x12.Records.ShareUser\x12\x36\n\x14shareEnterpriseNames\x18\x07 \x03(\x0b\x32\x18.Records.ShareEnterprise\"\xa5\x01\n\tShareUser\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x10\n\x08\x66ullname\x18\x02 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x03 \x01(\x05\x12$\n\x06status\x18\x04 \x01(\x0e\x32\x14.Records.ShareStatus\x12\x14\n\x0cisShareAdmin\x18\x05 \x01(\x08\x12\"\n\x1aisAdminOfSharedFolderOwner\x18\x06 \x01(\x08\"D\n\tShareTeam\x12\x10\n\x08teamname\x18\x01 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x02 \x01(\x05\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\"?\n\x0fShareEnterprise\x12\x16\n\x0e\x65nterprisename\x18\x01 \x01(\t\x12\x14\n\x0c\x65nterpriseId\x18\x02 \x01(\x05\"S\n\x1fRecordsOnwershipTransferRequest\x12\x30\n\x0ftransferRecords\x18\x01 \x03(\x0b\x32\x17.Records.TransferRecord\"[\n\x0eTransferRecord\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x11\n\trecordKey\x18\x03 \x01(\x0c\x12\x11\n\tuseEccKey\x18\x04 \x01(\x08\"_\n RecordsOnwershipTransferResponse\x12;\n\x14transferRecordStatus\x18\x01 \x03(\x0b\x32\x1d.Records.TransferRecordStatus\"\\\n\x14TransferRecordStatus\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x0e\n\x06status\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\t*B\n\x0fRecordTypeScope\x12\x0f\n\x0bRT_STANDARD\x10\x00\x12\x0b\n\x07RT_USER\x10\x01\x12\x11\n\rRT_ENTERPRISE\x10\x02*\x93\x01\n\rRecordKeyType\x12\n\n\x06NO_KEY\x10\x00\x12\x19\n\x15\x45NCRYPTED_BY_DATA_KEY\x10\x01\x12\x1b\n\x17\x45NCRYPTED_BY_PUBLIC_KEY\x10\x02\x12\x1d\n\x19\x45NCRYPTED_BY_DATA_KEY_GCM\x10\x03\x12\x1f\n\x1b\x45NCRYPTED_BY_PUBLIC_KEY_ECC\x10\x04*P\n\x10RecordFolderType\x12\x0f\n\x0buser_folder\x10\x00\x12\x11\n\rshared_folder\x10\x01\x12\x18\n\x14shared_folder_folder\x10\x02*\xaa\x02\n\x12RecordModifyResult\x12\x0e\n\nRS_SUCCESS\x10\x00\x12\x12\n\x0eRS_OUT_OF_SYNC\x10\x01\x12\x14\n\x10RS_ACCESS_DENIED\x10\x02\x12\x13\n\x0fRS_SHARE_DENIED\x10\x03\x12\x14\n\x10RS_RECORD_EXISTS\x10\x04\x12\x1e\n\x1aRS_OLD_RECORD_VERSION_TYPE\x10\x05\x12\x1e\n\x1aRS_NEW_RECORD_VERSION_TYPE\x10\x06\x12\x16\n\x12RS_FILES_NOT_MATCH\x10\x07\x12\x1b\n\x17RS_RECORD_NOT_SHAREABLE\x10\x08\x12\x1f\n\x1bRS_ATTACHMENT_NOT_SHAREABLE\x10\t\x12\x19\n\x15RS_FILE_LIMIT_REACHED\x10\n*-\n\rFileAddResult\x12\x0e\n\nFA_SUCCESS\x10\x00\x12\x0c\n\x08\x46\x41_ERROR\x10\x01*C\n\rFileGetResult\x12\x0e\n\nFG_SUCCESS\x10\x00\x12\x0c\n\x08\x46G_ERROR\x10\x01\x12\x14\n\x10\x46G_ACCESS_DENIED\x10\x02*J\n\x14RecordDetailsInclude\x12\x13\n\x0f\x44\x41TA_PLUS_SHARE\x10\x00\x12\r\n\tDATA_ONLY\x10\x01\x12\x0e\n\nSHARE_ONLY\x10\x02*b\n\x19\x43heckShareAdminObjectType\x12\x19\n\x15\x43HECK_SA_INVALID_TYPE\x10\x00\x12\x12\n\x0e\x43HECK_SA_ON_SF\x10\x01\x12\x16\n\x12\x43HECK_SA_ON_RECORD\x10\x02*1\n\x0bShareStatus\x12\n\n\x06\x41\x43TIVE\x10\x00\x12\t\n\x05\x42LOCK\x10\x01\x12\x0b\n\x07INVITED\x10\x02\x42#\n\x18\x63om.keepersecurity.protoB\x07Recordsb\x06proto3')

_RECORDTYPESCOPE = DESCRIPTOR.enum_types_by_name['RecordTypeScope']
RecordTypeScope = enum_type_wrapper.EnumTypeWrapper(_RECORDTYPESCOPE)
_RECORDKEYTYPE = DESCRIPTOR.enum_types_by_name['RecordKeyType']
RecordKeyType = enum_type_wrapper.EnumTypeWrapper(_RECORDKEYTYPE)
_RECORDFOLDERTYPE = DESCRIPTOR.enum_types_by_name['RecordFolderType']
RecordFolderType = enum_type_wrapper.EnumTypeWrapper(_RECORDFOLDERTYPE)
_RECORDMODIFYRESULT = DESCRIPTOR.enum_types_by_name['RecordModifyResult']
RecordModifyResult = enum_type_wrapper.EnumTypeWrapper(_RECORDMODIFYRESULT)
_FILEADDRESULT = DESCRIPTOR.enum_types_by_name['FileAddResult']
FileAddResult = enum_type_wrapper.EnumTypeWrapper(_FILEADDRESULT)
_FILEGETRESULT = DESCRIPTOR.enum_types_by_name['FileGetResult']
FileGetResult = enum_type_wrapper.EnumTypeWrapper(_FILEGETRESULT)
_RECORDDETAILSINCLUDE = DESCRIPTOR.enum_types_by_name['RecordDetailsInclude']
RecordDetailsInclude = enum_type_wrapper.EnumTypeWrapper(_RECORDDETAILSINCLUDE)
_CHECKSHAREADMINOBJECTTYPE = DESCRIPTOR.enum_types_by_name['CheckShareAdminObjectType']
CheckShareAdminObjectType = enum_type_wrapper.EnumTypeWrapper(_CHECKSHAREADMINOBJECTTYPE)
_SHARESTATUS = DESCRIPTOR.enum_types_by_name['ShareStatus']
ShareStatus = enum_type_wrapper.EnumTypeWrapper(_SHARESTATUS)
RT_STANDARD = 0
RT_USER = 1
RT_ENTERPRISE = 2
NO_KEY = 0
ENCRYPTED_BY_DATA_KEY = 1
ENCRYPTED_BY_PUBLIC_KEY = 2
ENCRYPTED_BY_DATA_KEY_GCM = 3
ENCRYPTED_BY_PUBLIC_KEY_ECC = 4
user_folder = 0
shared_folder = 1
shared_folder_folder = 2
RS_SUCCESS = 0
RS_OUT_OF_SYNC = 1
RS_ACCESS_DENIED = 2
RS_SHARE_DENIED = 3
RS_RECORD_EXISTS = 4
RS_OLD_RECORD_VERSION_TYPE = 5
RS_NEW_RECORD_VERSION_TYPE = 6
RS_FILES_NOT_MATCH = 7
RS_RECORD_NOT_SHAREABLE = 8
RS_ATTACHMENT_NOT_SHAREABLE = 9
RS_FILE_LIMIT_REACHED = 10
FA_SUCCESS = 0
FA_ERROR = 1
FG_SUCCESS = 0
FG_ERROR = 1
FG_ACCESS_DENIED = 2
DATA_PLUS_SHARE = 0
DATA_ONLY = 1
SHARE_ONLY = 2
CHECK_SA_INVALID_TYPE = 0
CHECK_SA_ON_SF = 1
CHECK_SA_ON_RECORD = 2
ACTIVE = 0
BLOCK = 1
INVITED = 2


_RECORDTYPE = DESCRIPTOR.message_types_by_name['RecordType']
_RECORDTYPESREQUEST = DESCRIPTOR.message_types_by_name['RecordTypesRequest']
_RECORDTYPESRESPONSE = DESCRIPTOR.message_types_by_name['RecordTypesResponse']
_RECORDTYPEMODIFYRESPONSE = DESCRIPTOR.message_types_by_name['RecordTypeModifyResponse']
_RECORDSGETREQUEST = DESCRIPTOR.message_types_by_name['RecordsGetRequest']
_RECORD = DESCRIPTOR.message_types_by_name['Record']
_FOLDERRECORDKEY = DESCRIPTOR.message_types_by_name['FolderRecordKey']
_FOLDER = DESCRIPTOR.message_types_by_name['Folder']
_TEAM = DESCRIPTOR.message_types_by_name['Team']
_RECORDSGETRESPONSE = DESCRIPTOR.message_types_by_name['RecordsGetResponse']
_RECORDLINK = DESCRIPTOR.message_types_by_name['RecordLink']
_RECORDAUDIT = DESCRIPTOR.message_types_by_name['RecordAudit']
_RECORDADD = DESCRIPTOR.message_types_by_name['RecordAdd']
_RECORDSADDREQUEST = DESCRIPTOR.message_types_by_name['RecordsAddRequest']
_RECORDUPDATE = DESCRIPTOR.message_types_by_name['RecordUpdate']
_RECORDSUPDATEREQUEST = DESCRIPTOR.message_types_by_name['RecordsUpdateRequest']
_RECORDFILEFORCONVERSION = DESCRIPTOR.message_types_by_name['RecordFileForConversion']
_RECORDFOLDERFORCONVERSION = DESCRIPTOR.message_types_by_name['RecordFolderForConversion']
_RECORDCONVERTTOV3 = DESCRIPTOR.message_types_by_name['RecordConvertToV3']
_RECORDSCONVERTTOV3REQUEST = DESCRIPTOR.message_types_by_name['RecordsConvertToV3Request']
_RECORDSREMOVEREQUEST = DESCRIPTOR.message_types_by_name['RecordsRemoveRequest']
_RECORDREVERT = DESCRIPTOR.message_types_by_name['RecordRevert']
_RECORDSREVERTREQUEST = DESCRIPTOR.message_types_by_name['RecordsRevertRequest']
_RECORDLINKERROR = DESCRIPTOR.message_types_by_name['RecordLinkError']
_RECORDMODIFYSTATUS = DESCRIPTOR.message_types_by_name['RecordModifyStatus']
_RECORDSMODIFYRESPONSE = DESCRIPTOR.message_types_by_name['RecordsModifyResponse']
_RECORDADDAUDITDATA = DESCRIPTOR.message_types_by_name['RecordAddAuditData']
_ADDAUDITDATAREQUEST = DESCRIPTOR.message_types_by_name['AddAuditDataRequest']
_FILE = DESCRIPTOR.message_types_by_name['File']
_FILESADDREQUEST = DESCRIPTOR.message_types_by_name['FilesAddRequest']
_FILEADDSTATUS = DESCRIPTOR.message_types_by_name['FileAddStatus']
_FILESADDRESPONSE = DESCRIPTOR.message_types_by_name['FilesAddResponse']
_FILESGETREQUEST = DESCRIPTOR.message_types_by_name['FilesGetRequest']
_FILEGETSTATUS = DESCRIPTOR.message_types_by_name['FileGetStatus']
_FILESGETRESPONSE = DESCRIPTOR.message_types_by_name['FilesGetResponse']
_APPLICATIONADDREQUEST = DESCRIPTOR.message_types_by_name['ApplicationAddRequest']
_GETRECORDDATAWITHACCESSINFOREQUEST = DESCRIPTOR.message_types_by_name['GetRecordDataWithAccessInfoRequest']
_USERPERMISSION = DESCRIPTOR.message_types_by_name['UserPermission']
_SHAREDFOLDERPERMISSION = DESCRIPTOR.message_types_by_name['SharedFolderPermission']
_RECORDDATA = DESCRIPTOR.message_types_by_name['RecordData']
_RECORDDATAWITHACCESSINFO = DESCRIPTOR.message_types_by_name['RecordDataWithAccessInfo']
_GETRECORDDATAWITHACCESSINFORESPONSE = DESCRIPTOR.message_types_by_name['GetRecordDataWithAccessInfoResponse']
_ISOBJECTSHAREADMIN = DESCRIPTOR.message_types_by_name['IsObjectShareAdmin']
_AMISHAREADMIN = DESCRIPTOR.message_types_by_name['AmIShareAdmin']
_RECORDSHAREUPDATEREQUEST = DESCRIPTOR.message_types_by_name['RecordShareUpdateRequest']
_SHAREDRECORD = DESCRIPTOR.message_types_by_name['SharedRecord']
_RECORDSHAREUPDATERESPONSE = DESCRIPTOR.message_types_by_name['RecordShareUpdateResponse']
_SHAREDRECORDSTATUS = DESCRIPTOR.message_types_by_name['SharedRecordStatus']
_GETRECORDPERMISSIONSREQUEST = DESCRIPTOR.message_types_by_name['GetRecordPermissionsRequest']
_GETRECORDPERMISSIONSRESPONSE = DESCRIPTOR.message_types_by_name['GetRecordPermissionsResponse']
_RECORDPERMISSION = DESCRIPTOR.message_types_by_name['RecordPermission']
_GETSHAREOBJECTSREQUEST = DESCRIPTOR.message_types_by_name['GetShareObjectsRequest']
_GETSHAREOBJECTSRESPONSE = DESCRIPTOR.message_types_by_name['GetShareObjectsResponse']
_SHAREUSER = DESCRIPTOR.message_types_by_name['ShareUser']
_SHARETEAM = DESCRIPTOR.message_types_by_name['ShareTeam']
_SHAREENTERPRISE = DESCRIPTOR.message_types_by_name['ShareEnterprise']
_RECORDSONWERSHIPTRANSFERREQUEST = DESCRIPTOR.message_types_by_name['RecordsOnwershipTransferRequest']
_TRANSFERRECORD = DESCRIPTOR.message_types_by_name['TransferRecord']
_RECORDSONWERSHIPTRANSFERRESPONSE = DESCRIPTOR.message_types_by_name['RecordsOnwershipTransferResponse']
_TRANSFERRECORDSTATUS = DESCRIPTOR.message_types_by_name['TransferRecordStatus']
RecordType = _reflection.GeneratedProtocolMessageType('RecordType', (_message.Message,), {
  'DESCRIPTOR' : _RECORDTYPE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordType)
  })
_sym_db.RegisterMessage(RecordType)

RecordTypesRequest = _reflection.GeneratedProtocolMessageType('RecordTypesRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDTYPESREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordTypesRequest)
  })
_sym_db.RegisterMessage(RecordTypesRequest)

RecordTypesResponse = _reflection.GeneratedProtocolMessageType('RecordTypesResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDTYPESRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordTypesResponse)
  })
_sym_db.RegisterMessage(RecordTypesResponse)

RecordTypeModifyResponse = _reflection.GeneratedProtocolMessageType('RecordTypeModifyResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDTYPEMODIFYRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordTypeModifyResponse)
  })
_sym_db.RegisterMessage(RecordTypeModifyResponse)

RecordsGetRequest = _reflection.GeneratedProtocolMessageType('RecordsGetRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSGETREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsGetRequest)
  })
_sym_db.RegisterMessage(RecordsGetRequest)

Record = _reflection.GeneratedProtocolMessageType('Record', (_message.Message,), {
  'DESCRIPTOR' : _RECORD,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.Record)
  })
_sym_db.RegisterMessage(Record)

FolderRecordKey = _reflection.GeneratedProtocolMessageType('FolderRecordKey', (_message.Message,), {
  'DESCRIPTOR' : _FOLDERRECORDKEY,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FolderRecordKey)
  })
_sym_db.RegisterMessage(FolderRecordKey)

Folder = _reflection.GeneratedProtocolMessageType('Folder', (_message.Message,), {
  'DESCRIPTOR' : _FOLDER,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.Folder)
  })
_sym_db.RegisterMessage(Folder)

Team = _reflection.GeneratedProtocolMessageType('Team', (_message.Message,), {
  'DESCRIPTOR' : _TEAM,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.Team)
  })
_sym_db.RegisterMessage(Team)

RecordsGetResponse = _reflection.GeneratedProtocolMessageType('RecordsGetResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSGETRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsGetResponse)
  })
_sym_db.RegisterMessage(RecordsGetResponse)

RecordLink = _reflection.GeneratedProtocolMessageType('RecordLink', (_message.Message,), {
  'DESCRIPTOR' : _RECORDLINK,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordLink)
  })
_sym_db.RegisterMessage(RecordLink)

RecordAudit = _reflection.GeneratedProtocolMessageType('RecordAudit', (_message.Message,), {
  'DESCRIPTOR' : _RECORDAUDIT,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordAudit)
  })
_sym_db.RegisterMessage(RecordAudit)

RecordAdd = _reflection.GeneratedProtocolMessageType('RecordAdd', (_message.Message,), {
  'DESCRIPTOR' : _RECORDADD,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordAdd)
  })
_sym_db.RegisterMessage(RecordAdd)

RecordsAddRequest = _reflection.GeneratedProtocolMessageType('RecordsAddRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSADDREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsAddRequest)
  })
_sym_db.RegisterMessage(RecordsAddRequest)

RecordUpdate = _reflection.GeneratedProtocolMessageType('RecordUpdate', (_message.Message,), {
  'DESCRIPTOR' : _RECORDUPDATE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordUpdate)
  })
_sym_db.RegisterMessage(RecordUpdate)

RecordsUpdateRequest = _reflection.GeneratedProtocolMessageType('RecordsUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSUPDATEREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsUpdateRequest)
  })
_sym_db.RegisterMessage(RecordsUpdateRequest)

RecordFileForConversion = _reflection.GeneratedProtocolMessageType('RecordFileForConversion', (_message.Message,), {
  'DESCRIPTOR' : _RECORDFILEFORCONVERSION,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordFileForConversion)
  })
_sym_db.RegisterMessage(RecordFileForConversion)

RecordFolderForConversion = _reflection.GeneratedProtocolMessageType('RecordFolderForConversion', (_message.Message,), {
  'DESCRIPTOR' : _RECORDFOLDERFORCONVERSION,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordFolderForConversion)
  })
_sym_db.RegisterMessage(RecordFolderForConversion)

RecordConvertToV3 = _reflection.GeneratedProtocolMessageType('RecordConvertToV3', (_message.Message,), {
  'DESCRIPTOR' : _RECORDCONVERTTOV3,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordConvertToV3)
  })
_sym_db.RegisterMessage(RecordConvertToV3)

RecordsConvertToV3Request = _reflection.GeneratedProtocolMessageType('RecordsConvertToV3Request', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSCONVERTTOV3REQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsConvertToV3Request)
  })
_sym_db.RegisterMessage(RecordsConvertToV3Request)

RecordsRemoveRequest = _reflection.GeneratedProtocolMessageType('RecordsRemoveRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSREMOVEREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsRemoveRequest)
  })
_sym_db.RegisterMessage(RecordsRemoveRequest)

RecordRevert = _reflection.GeneratedProtocolMessageType('RecordRevert', (_message.Message,), {
  'DESCRIPTOR' : _RECORDREVERT,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordRevert)
  })
_sym_db.RegisterMessage(RecordRevert)

RecordsRevertRequest = _reflection.GeneratedProtocolMessageType('RecordsRevertRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSREVERTREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsRevertRequest)
  })
_sym_db.RegisterMessage(RecordsRevertRequest)

RecordLinkError = _reflection.GeneratedProtocolMessageType('RecordLinkError', (_message.Message,), {
  'DESCRIPTOR' : _RECORDLINKERROR,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordLinkError)
  })
_sym_db.RegisterMessage(RecordLinkError)

RecordModifyStatus = _reflection.GeneratedProtocolMessageType('RecordModifyStatus', (_message.Message,), {
  'DESCRIPTOR' : _RECORDMODIFYSTATUS,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordModifyStatus)
  })
_sym_db.RegisterMessage(RecordModifyStatus)

RecordsModifyResponse = _reflection.GeneratedProtocolMessageType('RecordsModifyResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSMODIFYRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsModifyResponse)
  })
_sym_db.RegisterMessage(RecordsModifyResponse)

RecordAddAuditData = _reflection.GeneratedProtocolMessageType('RecordAddAuditData', (_message.Message,), {
  'DESCRIPTOR' : _RECORDADDAUDITDATA,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordAddAuditData)
  })
_sym_db.RegisterMessage(RecordAddAuditData)

AddAuditDataRequest = _reflection.GeneratedProtocolMessageType('AddAuditDataRequest', (_message.Message,), {
  'DESCRIPTOR' : _ADDAUDITDATAREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.AddAuditDataRequest)
  })
_sym_db.RegisterMessage(AddAuditDataRequest)

File = _reflection.GeneratedProtocolMessageType('File', (_message.Message,), {
  'DESCRIPTOR' : _FILE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.File)
  })
_sym_db.RegisterMessage(File)

FilesAddRequest = _reflection.GeneratedProtocolMessageType('FilesAddRequest', (_message.Message,), {
  'DESCRIPTOR' : _FILESADDREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FilesAddRequest)
  })
_sym_db.RegisterMessage(FilesAddRequest)

FileAddStatus = _reflection.GeneratedProtocolMessageType('FileAddStatus', (_message.Message,), {
  'DESCRIPTOR' : _FILEADDSTATUS,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FileAddStatus)
  })
_sym_db.RegisterMessage(FileAddStatus)

FilesAddResponse = _reflection.GeneratedProtocolMessageType('FilesAddResponse', (_message.Message,), {
  'DESCRIPTOR' : _FILESADDRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FilesAddResponse)
  })
_sym_db.RegisterMessage(FilesAddResponse)

FilesGetRequest = _reflection.GeneratedProtocolMessageType('FilesGetRequest', (_message.Message,), {
  'DESCRIPTOR' : _FILESGETREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FilesGetRequest)
  })
_sym_db.RegisterMessage(FilesGetRequest)

FileGetStatus = _reflection.GeneratedProtocolMessageType('FileGetStatus', (_message.Message,), {
  'DESCRIPTOR' : _FILEGETSTATUS,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FileGetStatus)
  })
_sym_db.RegisterMessage(FileGetStatus)

FilesGetResponse = _reflection.GeneratedProtocolMessageType('FilesGetResponse', (_message.Message,), {
  'DESCRIPTOR' : _FILESGETRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.FilesGetResponse)
  })
_sym_db.RegisterMessage(FilesGetResponse)

ApplicationAddRequest = _reflection.GeneratedProtocolMessageType('ApplicationAddRequest', (_message.Message,), {
  'DESCRIPTOR' : _APPLICATIONADDREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.ApplicationAddRequest)
  })
_sym_db.RegisterMessage(ApplicationAddRequest)

GetRecordDataWithAccessInfoRequest = _reflection.GeneratedProtocolMessageType('GetRecordDataWithAccessInfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRECORDDATAWITHACCESSINFOREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetRecordDataWithAccessInfoRequest)
  })
_sym_db.RegisterMessage(GetRecordDataWithAccessInfoRequest)

UserPermission = _reflection.GeneratedProtocolMessageType('UserPermission', (_message.Message,), {
  'DESCRIPTOR' : _USERPERMISSION,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.UserPermission)
  })
_sym_db.RegisterMessage(UserPermission)

SharedFolderPermission = _reflection.GeneratedProtocolMessageType('SharedFolderPermission', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERPERMISSION,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.SharedFolderPermission)
  })
_sym_db.RegisterMessage(SharedFolderPermission)

RecordData = _reflection.GeneratedProtocolMessageType('RecordData', (_message.Message,), {
  'DESCRIPTOR' : _RECORDDATA,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordData)
  })
_sym_db.RegisterMessage(RecordData)

RecordDataWithAccessInfo = _reflection.GeneratedProtocolMessageType('RecordDataWithAccessInfo', (_message.Message,), {
  'DESCRIPTOR' : _RECORDDATAWITHACCESSINFO,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordDataWithAccessInfo)
  })
_sym_db.RegisterMessage(RecordDataWithAccessInfo)

GetRecordDataWithAccessInfoResponse = _reflection.GeneratedProtocolMessageType('GetRecordDataWithAccessInfoResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETRECORDDATAWITHACCESSINFORESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetRecordDataWithAccessInfoResponse)
  })
_sym_db.RegisterMessage(GetRecordDataWithAccessInfoResponse)

IsObjectShareAdmin = _reflection.GeneratedProtocolMessageType('IsObjectShareAdmin', (_message.Message,), {
  'DESCRIPTOR' : _ISOBJECTSHAREADMIN,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.IsObjectShareAdmin)
  })
_sym_db.RegisterMessage(IsObjectShareAdmin)

AmIShareAdmin = _reflection.GeneratedProtocolMessageType('AmIShareAdmin', (_message.Message,), {
  'DESCRIPTOR' : _AMISHAREADMIN,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.AmIShareAdmin)
  })
_sym_db.RegisterMessage(AmIShareAdmin)

RecordShareUpdateRequest = _reflection.GeneratedProtocolMessageType('RecordShareUpdateRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSHAREUPDATEREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordShareUpdateRequest)
  })
_sym_db.RegisterMessage(RecordShareUpdateRequest)

SharedRecord = _reflection.GeneratedProtocolMessageType('SharedRecord', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDRECORD,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.SharedRecord)
  })
_sym_db.RegisterMessage(SharedRecord)

RecordShareUpdateResponse = _reflection.GeneratedProtocolMessageType('RecordShareUpdateResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSHAREUPDATERESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordShareUpdateResponse)
  })
_sym_db.RegisterMessage(RecordShareUpdateResponse)

SharedRecordStatus = _reflection.GeneratedProtocolMessageType('SharedRecordStatus', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDRECORDSTATUS,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.SharedRecordStatus)
  })
_sym_db.RegisterMessage(SharedRecordStatus)

GetRecordPermissionsRequest = _reflection.GeneratedProtocolMessageType('GetRecordPermissionsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETRECORDPERMISSIONSREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetRecordPermissionsRequest)
  })
_sym_db.RegisterMessage(GetRecordPermissionsRequest)

GetRecordPermissionsResponse = _reflection.GeneratedProtocolMessageType('GetRecordPermissionsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETRECORDPERMISSIONSRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetRecordPermissionsResponse)
  })
_sym_db.RegisterMessage(GetRecordPermissionsResponse)

RecordPermission = _reflection.GeneratedProtocolMessageType('RecordPermission', (_message.Message,), {
  'DESCRIPTOR' : _RECORDPERMISSION,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordPermission)
  })
_sym_db.RegisterMessage(RecordPermission)

GetShareObjectsRequest = _reflection.GeneratedProtocolMessageType('GetShareObjectsRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETSHAREOBJECTSREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetShareObjectsRequest)
  })
_sym_db.RegisterMessage(GetShareObjectsRequest)

GetShareObjectsResponse = _reflection.GeneratedProtocolMessageType('GetShareObjectsResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETSHAREOBJECTSRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.GetShareObjectsResponse)
  })
_sym_db.RegisterMessage(GetShareObjectsResponse)

ShareUser = _reflection.GeneratedProtocolMessageType('ShareUser', (_message.Message,), {
  'DESCRIPTOR' : _SHAREUSER,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.ShareUser)
  })
_sym_db.RegisterMessage(ShareUser)

ShareTeam = _reflection.GeneratedProtocolMessageType('ShareTeam', (_message.Message,), {
  'DESCRIPTOR' : _SHARETEAM,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.ShareTeam)
  })
_sym_db.RegisterMessage(ShareTeam)

ShareEnterprise = _reflection.GeneratedProtocolMessageType('ShareEnterprise', (_message.Message,), {
  'DESCRIPTOR' : _SHAREENTERPRISE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.ShareEnterprise)
  })
_sym_db.RegisterMessage(ShareEnterprise)

RecordsOnwershipTransferRequest = _reflection.GeneratedProtocolMessageType('RecordsOnwershipTransferRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSONWERSHIPTRANSFERREQUEST,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsOnwershipTransferRequest)
  })
_sym_db.RegisterMessage(RecordsOnwershipTransferRequest)

TransferRecord = _reflection.GeneratedProtocolMessageType('TransferRecord', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFERRECORD,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.TransferRecord)
  })
_sym_db.RegisterMessage(TransferRecord)

RecordsOnwershipTransferResponse = _reflection.GeneratedProtocolMessageType('RecordsOnwershipTransferResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDSONWERSHIPTRANSFERRESPONSE,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.RecordsOnwershipTransferResponse)
  })
_sym_db.RegisterMessage(RecordsOnwershipTransferResponse)

TransferRecordStatus = _reflection.GeneratedProtocolMessageType('TransferRecordStatus', (_message.Message,), {
  'DESCRIPTOR' : _TRANSFERRECORDSTATUS,
  '__module__' : 'record_pb2'
  # @@protoc_insertion_point(class_scope:Records.TransferRecordStatus)
  })
_sym_db.RegisterMessage(TransferRecordStatus)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\007Records'
  _RECORDTYPESCOPE._serialized_start=7537
  _RECORDTYPESCOPE._serialized_end=7603
  _RECORDKEYTYPE._serialized_start=7606
  _RECORDKEYTYPE._serialized_end=7753
  _RECORDFOLDERTYPE._serialized_start=7755
  _RECORDFOLDERTYPE._serialized_end=7835
  _RECORDMODIFYRESULT._serialized_start=7838
  _RECORDMODIFYRESULT._serialized_end=8136
  _FILEADDRESULT._serialized_start=8138
  _FILEADDRESULT._serialized_end=8183
  _FILEGETRESULT._serialized_start=8185
  _FILEGETRESULT._serialized_end=8252
  _RECORDDETAILSINCLUDE._serialized_start=8254
  _RECORDDETAILSINCLUDE._serialized_end=8328
  _CHECKSHAREADMINOBJECTTYPE._serialized_start=8330
  _CHECKSHAREADMINOBJECTTYPE._serialized_end=8428
  _SHARESTATUS._serialized_start=8430
  _SHARESTATUS._serialized_end=8479
  _RECORDTYPE._serialized_start=25
  _RECORDTYPE._serialized_end=117
  _RECORDTYPESREQUEST._serialized_start=119
  _RECORDTYPESREQUEST._serialized_end=191
  _RECORDTYPESRESPONSE._serialized_start=194
  _RECORDTYPESRESPONSE._serialized_end=330
  _RECORDTYPEMODIFYRESPONSE._serialized_start=332
  _RECORDTYPEMODIFYRESPONSE._serialized_end=397
  _RECORDSGETREQUEST._serialized_start=399
  _RECORDSGETREQUEST._serialized_end=460
  _RECORD._serialized_start=463
  _RECORD._serialized_end=672
  _FOLDERRECORDKEY._serialized_start=674
  _FOLDERRECORDKEY._serialized_end=751
  _FOLDER._serialized_start=753
  _FOLDER._serialized_end=850
  _TEAM._serialized_start=853
  _TEAM._serialized_end=1002
  _RECORDSGETRESPONSE._serialized_start=1005
  _RECORDSGETRESPONSE._serialized_end=1177
  _RECORDLINK._serialized_start=1179
  _RECORDLINK._serialized_end=1231
  _RECORDAUDIT._serialized_start=1233
  _RECORDAUDIT._serialized_end=1277
  _RECORDADD._serialized_start=1280
  _RECORDADD._serialized_end=1568
  _RECORDSADDREQUEST._serialized_start=1570
  _RECORDSADDREQUEST._serialized_end=1647
  _RECORDUPDATE._serialized_start=1650
  _RECORDUPDATE._serialized_end=1884
  _RECORDSUPDATEREQUEST._serialized_start=1886
  _RECORDSUPDATEREQUEST._serialized_end=1969
  _RECORDFILEFORCONVERSION._serialized_start=1972
  _RECORDFILEFORCONVERSION._serialized_end=2114
  _RECORDFOLDERFORCONVERSION._serialized_start=2116
  _RECORDFOLDERFORCONVERSION._serialized_end=2190
  _RECORDCONVERTTOV3._serialized_start=2193
  _RECORDCONVERTTOV3._serialized_end=2467
  _RECORDSCONVERTTOV3REQUEST._serialized_start=2469
  _RECORDSCONVERTTOV3REQUEST._serialized_end=2562
  _RECORDSREMOVEREQUEST._serialized_start=2564
  _RECORDSREMOVEREQUEST._serialized_end=2603
  _RECORDREVERT._serialized_start=2605
  _RECORDREVERT._serialized_end=2667
  _RECORDSREVERTREQUEST._serialized_start=2669
  _RECORDSREVERTREQUEST._serialized_end=2731
  _RECORDLINKERROR._serialized_start=2733
  _RECORDLINKERROR._serialized_end=2832
  _RECORDMODIFYSTATUS._serialized_start=2835
  _RECORDMODIFYSTATUS._serialized_end=2984
  _RECORDSMODIFYRESPONSE._serialized_start=2986
  _RECORDSMODIFYRESPONSE._serialized_end=3073
  _RECORDADDAUDITDATA._serialized_start=3075
  _RECORDADDAUDITDATA._serialized_end=3164
  _ADDAUDITDATAREQUEST._serialized_start=3166
  _ADDAUDITDATAREQUEST._serialized_end=3233
  _FILE._serialized_start=3235
  _FILE._serialized_end=3332
  _FILESADDREQUEST._serialized_start=3334
  _FILESADDREQUEST._serialized_end=3402
  _FILEADDSTATUS._serialized_start=3405
  _FILEADDSTATUS._serialized_end=3572
  _FILESADDRESPONSE._serialized_start=3574
  _FILESADDRESPONSE._serialized_end=3649
  _FILESGETREQUEST._serialized_start=3651
  _FILESGETREQUEST._serialized_end=3753
  _FILEGETSTATUS._serialized_start=3756
  _FILEGETSTATUS._serialized_end=3918
  _FILESGETRESPONSE._serialized_start=3920
  _FILESGETRESPONSE._serialized_end=3977
  _APPLICATIONADDREQUEST._serialized_start=3979
  _APPLICATIONADDREQUEST._serialized_end=4083
  _GETRECORDDATAWITHACCESSINFOREQUEST._serialized_start=4086
  _GETRECORDDATAWITHACCESSINFOREQUEST._serialized_end=4222
  _USERPERMISSION._serialized_start=4225
  _USERPERMISSION._serialized_end=4396
  _SHAREDFOLDERPERMISSION._serialized_start=4398
  _SHAREDFOLDERPERMISSION._serialized_end=4523
  _RECORDDATA._serialized_start=4526
  _RECORDDATA._serialized_end=4886
  _RECORDDATAWITHACCESSINFO._serialized_start=4889
  _RECORDDATAWITHACCESSINFO._serialized_end=5089
  _GETRECORDDATAWITHACCESSINFORESPONSE._serialized_start=5092
  _GETRECORDDATAWITHACCESSINFORESPONSE._serialized_end=5229
  _ISOBJECTSHAREADMIN._serialized_start=5231
  _ISOBJECTSHAREADMIN._serialized_end=5337
  _AMISHAREADMIN._serialized_start=5339
  _AMISHAREADMIN._serialized_end=5411
  _RECORDSHAREUPDATEREQUEST._serialized_start=5414
  _RECORDSHAREUPDATEREQUEST._serialized_end=5602
  _SHAREDRECORD._serialized_start=5605
  _SHAREDRECORD._serialized_end=5818
  _RECORDSHAREUPDATERESPONSE._serialized_start=5821
  _RECORDSHAREUPDATERESPONSE._serialized_end=6034
  _SHAREDRECORDSTATUS._serialized_start=6036
  _SHAREDRECORDSTATUS._serialized_end=6126
  _GETRECORDPERMISSIONSREQUEST._serialized_start=6128
  _GETRECORDPERMISSIONSREQUEST._serialized_end=6199
  _GETRECORDPERMISSIONSRESPONSE._serialized_start=6201
  _GETRECORDPERMISSIONSRESPONSE._serialized_end=6285
  _RECORDPERMISSION._serialized_start=6287
  _RECORDPERMISSION._serialized_end=6395
  _GETSHAREOBJECTSREQUEST._serialized_start=6397
  _GETSHAREOBJECTSREQUEST._serialized_end=6501
  _GETSHAREOBJECTSRESPONSE._serialized_start=6504
  _GETSHAREOBJECTSRESPONSE._serialized_end=6863
  _SHAREUSER._serialized_start=6866
  _SHAREUSER._serialized_end=7031
  _SHARETEAM._serialized_start=7033
  _SHARETEAM._serialized_end=7101
  _SHAREENTERPRISE._serialized_start=7103
  _SHAREENTERPRISE._serialized_end=7166
  _RECORDSONWERSHIPTRANSFERREQUEST._serialized_start=7168
  _RECORDSONWERSHIPTRANSFERREQUEST._serialized_end=7251
  _TRANSFERRECORD._serialized_start=7253
  _TRANSFERRECORD._serialized_end=7344
  _RECORDSONWERSHIPTRANSFERRESPONSE._serialized_start=7346
  _RECORDSONWERSHIPTRANSFERRESPONSE._serialized_end=7441
  _TRANSFERRECORDSTATUS._serialized_start=7443
  _TRANSFERRECORDSTATUS._serialized_end=7535
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import argparse
import logging
import time
from glob import glob
from os.path import normpath

from .. import cli
from .base import raise_parse_exception, suppress_exit, Command


def register_commands(commands):
    commands['run-batch'] = RunBatchCommand()
    commands['sleep'] = SleepCommand()


def register_command_info(aliases, command_info):
    aliases['run'] = 'run-batch'
    command_info[run_batch_parser.prog] = run_batch_parser.description
    command_info[sleep_parser.prog] = sleep_parser.description


run_batch_parser = argparse.ArgumentParser(prog='run-batch', description='Run batch of Commander commands from a file')
run_batch_parser.add_argument(
    '-d', '--delay', dest='delay', type=int, action='store',
    help='Delay (in seconds) between commands to prevent throttling'
)
run_batch_parser.add_argument(
    '-q', '--quiet', dest='quiet', action='store_true', help="Don't display batch file info"
)
run_batch_parser.add_argument(
    '-n', '--dry-run', dest='dry_run', action='store_true', help='Preview the commands that will be run'
)
run_batch_parser.add_argument(
    'batch-file-patterns', nargs='*', type=str, action='store', help='One or more batch files of Commander commands'
)
run_batch_parser.error = raise_parse_exception
run_batch_parser.exit = suppress_exit


sleep_parser = argparse.ArgumentParser(
    prog='sleep', description='Sleep (in seconds) for adding delay between batch commands'
)
sleep_parser.add_argument(
    'sleep-duration', nargs='?', type=int, action='store', help='Sleep duration in seconds'
)
sleep_parser.error = raise_parse_exception
sleep_parser.exit = suppress_exit


class RunBatchCommand(Command):
    def get_parser(self):
        return run_batch_parser

    def execute(self, params, **kwargs):
        dry_run = kwargs.get('dry_run', False)
        command_delay = kwargs.get('delay') or 0
        quiet = kwargs.get('quiet', False)
        pattern_list = kwargs.get('batch-file-patterns', [])
        if len(pattern_list) == 0:
            logging.warning(f'Please specify one or more batch files to run')
            return

        if dry_run:
            print('The following files and commands would be run:')
        for pattern in pattern_list:
            for filepath in glob(pattern):
                filepath = normpath(filepath)
                if dry_run:
                    print(f'{filepath}:')
                elif not quiet:
                    logging.info(f'Running Keeper Commander batch file {filepath}...')

                with open(filepath) as f:
                    lines = f.readlines()
                    commands = [c.strip() for c in lines if not c.startswith('#')]
                    if len(commands) > 0:
                        if dry_run:
                            print('    ' + '\n    '.join(commands))
                        else:
                            cli.runcommands(params, commands=commands, command_delay=command_delay, quiet=quiet)
                    else:
                        if dry_run:
                            print('No commands')
                        else:
                            logging.warning(f'No commands to execute in batch file {filepath}')


class SleepCommand(Command):
    def get_parser(self):
        return sleep_parser

    def execute(self, params, **kwargs):
        sleep_duration = kwargs.get('sleep-duration')
        if sleep_duration is None:
            logging.warning(f'Please specify the sleep duration in seconds')
            return
        time.sleep(sleep_duration)



import collections
import logging
import sqlite3
from typing import Dict, Union, Sequence, Any, List, Optional, Type, Callable, Iterable, Iterator

FieldSchema = collections.namedtuple('FieldSchema', ['name', 'type'])


class TableSchema:
    def __init__(self):
        self.table_name = ''            # type: str
        self.class_type = type(object)  # type: Type
        self.columns = []               # type: List[str]
        self.primary_key = []           # type: List[str]
        self.indexes = None             # type: Optional[Dict[str, List[str]]]
        self.owner_column = None        # type: Optional[str]
        self.owner_column_type = None   # type: Union[Type[int], Type[str], None]
        self.class_fields = {}          # type: Dict[str, FieldSchema]

    @classmethod
    def load_schema(cls, class_type, primary_key, indexes=None, owner_column=None, owner_type=None):
        # type: (Type, Sequence[str], Optional[Dict[str, Union[str, Sequence[str]]]], Optional[str], Optional[Type]) -> TableSchema
        schema = cls()
        schema.class_type = class_type
        schema.table_name = class_type.__name__
        obj_values = vars(class_type())
        for field_name, value in obj_values.items():
            column_name = field_name.lower()
            field_type = None   # type: Optional[Type]
            if isinstance(value, str):
                field_type = str
            elif isinstance(value, bool):
                field_type = bool
            elif isinstance(value, int):
                field_type = int
            elif isinstance(value, float):
                field_type = float
            elif isinstance(value, (bytes, bytearray)):
                field_type = bytes
            elif value is None:
                logging.debug(
                    'load_schema: Attribute \"%s\" in class \"%s\" is skipped since it is None',
                    field_name, schema.table_name)
            else:
                logging.debug(
                    'load_schema: Unsupported type for attribute \"%s\" in class \"%s\". Skipping',
                    field_name, schema.table_name)
            if field_type:
                schema.class_fields[column_name] = FieldSchema(field_name, field_type)
                schema.columns.append(field_name)

        if isinstance(primary_key, str):
            schema.primary_key.append(primary_key)
        elif isinstance(primary_key, (list, set)):
            schema.primary_key.extend(primary_key)
        else:
            if not owner_column:
                raise ValueError(f'Schema \"{schema.table_name}\" does not have either primary key or owner column')
        if len(schema.columns) == 0:
            raise ValueError(f'Table {schema.table_name} does not have primary key defined')

        for column in schema.primary_key:
            if column.lower() not in schema.class_fields:
                raise ValueError(f'Primary Key: Table {schema.table_name} does not have column {column}')

        if isinstance(indexes, dict):
            schema.indexes = {}
            for index_name, index_columns in indexes.items():
                if isinstance(index_columns, str):
                    index_columns = [index_columns]
                if not isinstance(index_columns, (list, tuple)):
                    raise ValueError(f'Index \"{index_name}\": invalid columns')
                for column in index_columns:
                    if column.lower() not in schema.class_fields:
                        raise ValueError(
                            f'Index \"{index_name}\": Table {schema.table_name} does not have column {column}')
                schema.indexes[index_name] = [x for x in index_columns]

        if owner_column:
            if owner_column.lower() in schema.class_fields:
                raise ValueError(f'Owner: Table {schema.table_name} contains owner column {owner_column}')
            schema.owner_column = owner_column
            schema.owner_column_type = str
            if owner_type and owner_type == int:
                schema.owner_column_type = int

        return schema


def _to_sqlite_type(column_type):
    # type: (Type) -> str
    if column_type in {bool, int}:
        return 'INTEGER'
    if column_type == float:
        return 'REAL'
    if column_type == bytes:
        return 'BLOB'
    return 'TEXT'


def verify_database(connection, tables, apply_changes=True):
    # type: (sqlite3.Connection, Sequence[TableSchema], bool) -> List[str]
    result = []
    existing_tables = set((x[0].lower() for x in
                           connection.execute('SELECT name FROM sqlite_master where type=?', ('table',))))
    for table in tables:
        queries = []
        if table.table_name.lower() in existing_tables:
            column_info = connection.execute(f'PRAGMA table_info(\'{table.table_name}\')').fetchall()
            column_info.sort(key=lambda x: x[0])
            columns = set((x[1].lower() for x in column_info))

            pk_cols = []
            if table.owner_column:
                pk_cols.append(table.owner_column)
            pk_cols.extend(table.primary_key)
            for col in pk_cols:
                if col.lower() not in columns:
                    raise ValueError(f'Table \"{table.table_name}\" misses primary key \"{col}\".')
            missing_columns = [x for x in table.columns if x.lower() not in columns]
            for col in missing_columns:
                field_schema = table.class_fields.get(col.lower())
                column_type = _to_sqlite_type(field_schema.type if field_schema else str)
                queries.append(f'ALTER TABLE {table.table_name} ADD COLUMN {col} {column_type}')

            if table.indexes:
                index_names = [x[1] for x in connection.execute(f'PRAGMA index_list(\'{table.table_name}\')')]
                indexes = {}
                for index_name in index_names:
                    index_info = connection.execute(f'PRAGMA index_info(\'{index_name}\')').fetchall()
                    index_info.sort(key=lambda x: x[0])
                    indexes[index_name.lower()] = [x[2] for x in index_info]
                for index_name, index_columns in table.indexes.items():
                    cols = []
                    if table.owner_column:
                        cols.append(table.owner_column)
                    cols.extend(index_columns)
                    index_found = None
                    for existing_index_name, existing_index_columns in indexes.items():
                        if len(existing_index_columns) != len(cols):
                            continue
                        if any(True for x in zip(existing_index_columns, cols) if x[0].lower() != x[1].lower()):
                            continue
                        index_found = existing_index_name
                        break
                    if index_found:
                        continue
                    index_name = f'{table.table_name}_{index_name}_IDX'
                    if index_name.lower() in indexes:
                        queries.append(f'DROP INDEX "{index_name}"')
                    queries.append(f'CREATE INDEX "{index_name}" ON {table.table_name} (' + ', '.join(cols) + ')')
        else:
            added_columns = set()
            table_columns = []
            pks = []
            if table.owner_column:
                pks.append(table.owner_column)
                column_type = _to_sqlite_type(table.owner_column_type or str)
                table_columns.append(f'\t{table.owner_column} {column_type} NOT NULL')
                added_columns.add(table.owner_column.lower())
            for pk_column in table.primary_key:
                pks.append(pk_column)
                column_type = _to_sqlite_type(table.class_fields[pk_column.lower()].type)
                table_columns.append(f'\t{pk_column} {column_type} NOT NULL')
                added_columns.add(pk_column.lower())
            for column in table.columns:
                if column.lower() in added_columns:
                    continue

                field_schema = table.class_fields.get(column.lower())
                column_type = _to_sqlite_type(field_schema.type if field_schema else str)
                table_columns.append(f'\t{column} {column_type}')
                added_columns.add(column.lower())
            queries.append(f'CREATE TABLE {table.table_name} (\n' + ',\n'.join(table_columns) + ',\n' +
                           '\tPRIMARY KEY (' + ', '.join(pks) + ')\n)')
            if table.indexes:
                for index_name, index_columns in table.indexes.items():
                    cols = []
                    if table.owner_column:
                        cols.append(table.owner_column)
                    cols.extend(index_columns)
                    queries.append(f'CREATE INDEX "{table.table_name}_{index_name}_IDX" '
                                   f'ON {table.table_name} (' + ', '.join(cols) + ')')

        if len(queries) > 0:
            for query in queries:
                if apply_changes:
                    connection.execute(query)
                else:
                    result.append(query)
            if apply_changes:
                connection.commit()
    return result


class SqliteStorage:
    def __init__(self, get_connection, schema, owner=None):
        # type: (Callable[[], sqlite3.Connection], TableSchema, Union[str, int, None]) -> None
        if not callable(get_connection):
            raise ValueError(f'"get_connection" should be callable.')
        self.get_connection = get_connection
        if not isinstance(schema, TableSchema):
            raise ValueError(f'"schema": Invalid type. TableSchema expected')
        self.schema = schema
        self.owner = None
        if owner:
            if not schema.owner_column:
                raise ValueError(f'"owner": schema does not define owner column.')
            if not isinstance(owner, schema.owner_column_type or str):
                raise ValueError(f'"owner": Invalid type. {(schema.owner_column_type or str).__name__} expected')
            self.owner = owner
        self._queries = {}    # type: Dict[str, str]

    def _populate_data_object(self, values):  # type: (Sequence[Any]) -> Any
        obj = self.schema.class_type()
        for i, column in enumerate(self.schema.columns):
            field_schema = self.schema.class_fields.get(column.lower())
            if field_schema:
                value = values[i]
                if field_schema.type == bool and isinstance(value, int):
                    value = value != 0
                setattr(obj, field_schema.name, value)
        return obj

    def _adjust_filter_columns(self, columns):  # type: (Union[str, Sequence[str]]) -> Sequence[str]
        if not columns:
            raise ValueError('adjust_filter_columns: columns cannot be empty')
        if isinstance(columns, str):
            columns = [columns]
        if not isinstance(columns, (list, tuple)):
            raise ValueError('adjust_filter_columns: columns should be a sequence of str')
        for column in columns:
            if column.lower() not in self.schema.class_fields:
                raise ValueError(
                    f'adjust_filter_columns: table \"{self.schema.table_name}\" does not have column \"{column}\"')
        return columns

    @staticmethod
    def _adjust_values_for_columns(columns, values):
        # type: (Sequence[str], Union[Any, Sequence[Any]]) -> Sequence[Any]
        if not isinstance(values, (list, tuple)):
            values = [values]
        if len(columns) != len(values):
            raise ValueError(
                f'adjust_values_for_columns: number of values {len(values)} does not match columns {len(columns)}')
        return values

    def prepare_params(self, columns, values):
        # type: (Sequence[str], Union[str, Sequence[Any]]) -> Dict[str, Any]
        params = {}
        if self.schema.owner_column:
            params[self.schema.owner_column] = self.owner
        adjusted_values = self._adjust_values_for_columns(columns, values)
        for i, column in enumerate(columns):
            params[column] = adjusted_values[i]
        return params

    def select_all(self, order_by=None):
        # type: (Union[str, Sequence[str], None]) -> Iterator[Any]
        key = 'select-all'
        if order_by:
            if isinstance(order_by, str):
                order_by = [order_by]
            elif isinstance(order_by, (list, tuple)):
                order_by = order_by
            else:
                raise ValueError(f'select_all: \"order_by\" invalid type.')
            for column_name in order_by:
                if not isinstance(column_name, str):
                    raise ValueError(f'select_all: \"order_by\" invalid type.')
                if column_name.lower() not in self.schema.class_fields:
                    raise ValueError(
                        f'select_all: table \"{self.schema.table_name}\" does not have column \"{column_name}\"')
            key += f': ' + ', '.join(order_by)

        query = self._queries.get(key)
        if not query:
            query = 'SELECT ' + ', '.join(self.schema.columns)
            query += f' FROM {self.schema.table_name}'
            if self.schema.owner_column:
                query += f' WHERE {self.schema.owner_column}=:{self.schema.owner_column}'
            if order_by:
                query += ' ORDER BY ' + ', '.join(order_by)
            self._queries[key] = query

        conn = self.get_connection()
        if self.schema.owner_column:
            params = {self.schema.owner_column: self.owner}
            curr = conn.execute(query, params)
        else:
            curr = conn.execute(query)
        for row in curr:
            yield self._populate_data_object(row)

    def select_by_filter(self, columns, values):
        # type: (Union[str, Sequence[str]], Union[Any, Sequence[Any]]) -> Iterable[Any]
        adjusted_columns = self._adjust_filter_columns(columns)

        key = f'select-by-filter: ' + ', '.join(adjusted_columns)
        query = self._queries.get(key)
        if not query:
            wheres = []
            if self.schema.owner_column:
                wheres.append(f'{self.schema.owner_column}=:{self.schema.owner_column}')
            wheres.extend((f'{x}=:{x}' for x in adjusted_columns))
            query = 'SELECT ' + ', '.join(self.schema.columns) + f' FROM {self.schema.table_name} ' +\
                    'WHERE ' + ' AND '.join(wheres)
            self._queries[key] = query

        conn = self.get_connection()
        curr = conn.execute(query, self.prepare_params(adjusted_columns, values))
        for row in curr:
            yield self._populate_data_object(row)

    def delete_all(self):   # type: () -> int
        query = self._queries.get('delete-all')
        if not query:
            wheres = []
            if self.schema.owner_column:
                wheres.append(f'{self.schema.owner_column}=:{self.schema.owner_column}')
            else:
                wheres.append('1=1')
            query = f'DELETE FROM {self.schema.table_name} WHERE ' + ' AND '.join(wheres)
            self._queries['delete-all'] = query

        conn = self.get_connection()
        try:
            if self.schema.owner_column:
                params = {self.schema.owner_column: self.owner}
                rs = conn.execute(query, params)
            else:
                rs = conn.execute(query)
            conn.commit()
            return rs.rowcount
        except Exception as e:
            conn.rollback()
            raise e

    def delete_by_filter(self, columns, values, multiple_criteria=False):
        # type: (Union[str, Sequence[str]], Union[Any, Sequence[Any]], bool) -> int
        adjusted_columns = self._adjust_filter_columns(columns)

        key = f'delete_by_filter: ' + ', '.join(adjusted_columns)
        query = self._queries.get(key)
        if not query:
            wheres = []
            if self.schema.owner_column:
                wheres.append(f'{self.schema.owner_column}=:{self.schema.owner_column}')
            wheres.extend((f'{x}=:{x}' for x in adjusted_columns))
            query = f'DELETE FROM {self.schema.table_name} WHERE ' + ' AND '.join(wheres)
            self._queries[key] = query

        conn = self.get_connection()
        row_count = 0
        try:
            if multiple_criteria:
                if not isinstance(values, (tuple, list)):
                    values = [values]
                rs = conn.executemany(query, (self.prepare_params(adjusted_columns, x) for x in values))
            else:
                rs = conn.execute(query, self.prepare_params(adjusted_columns, values))
            row_count += rs.rowcount
            conn.commit()
            return row_count
        except Exception as e:
            conn.rollback()
            raise e

    def get_entity_values(self, entity):   # type: (Any) -> Dict[str, Any]
        if not isinstance(entity, self.schema.class_type):
            raise ValueError('SqliteStorage:get_entity_values: invalid entity type. '
                             f'Expected {self.schema.class_type.__name__}')
        d = vars(entity)
        if self.schema.owner_column:
            d[self.schema.owner_column] = self.owner
        for column in self.schema.columns:
            if column not in d:
                d[column] = None
        return d

    def put(self, entities):       # type: (Union[Any, Sequence[Any]]) -> None
        key = 'put-entities'
        query = self._queries.get(key)
        if not query:
            cols = []
            if self.schema.owner_column:
                cols.append(self.schema.owner_column)
            cols.extend(self.schema.columns)
            query = f'INSERT OR REPLACE INTO {self.schema.table_name} (' + ', '.join(cols) + ') VALUES (' + \
                    ', '.join((f':{x}' for x in cols)) + ')'
            self._queries[key] = query

        conn = self.get_connection()
        try:
            conn.executemany(query, (self.get_entity_values(x) for x in entities))
            conn.commit()
        except Exception as e:
            conn.rollback()
            raise e



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import argparse
import base64
import datetime
import logging
import os
from urllib.parse import urlparse, urlunparse

from .base import user_choice, dump_report_data, GroupCommand
from .enterprise import EnterpriseCommand
from .. import api
from ..display import bcolors
from ..params import KeeperParams

scim_list_parser = argparse.ArgumentParser(prog='scim-list', description='Display a list of available SCIM endpoints.')

scim_view_parser = argparse.ArgumentParser(prog='scim-view', description='Display a SCIM endpoint details.')
scim_view_parser.add_argument('target', help='SCIM ID')

scim_create_parser = argparse.ArgumentParser(prog='scim-create', description='Create SCIM endpoint.')
scim_create_parser.add_argument('--node', dest='node', required=True, help='Node Name or ID.')
scim_create_parser.add_argument(
    '--prefix', dest='prefix', action='store',
    help='Role Prefix. SCIM groups staring with prefix will be imported to Keeper as Roles.')
scim_create_parser.add_argument(
    '--unique-groups', dest='unique_groups', action='store', choices=['on', 'off'], help='Unique Groups.')

scim_edit_parser = argparse.ArgumentParser(prog='scim-edit', description='Edit SCIM endpoint.')
scim_edit_parser.add_argument('target', help='SCIM ID')
scim_edit_parser.add_argument(
    '--prefix', dest='prefix', action='store',
    help='Role Prefix. SCIM groups staring with prefix will be imported to Keeper as Roles.')
scim_edit_parser.add_argument(
    '--unique-groups', dest='unique_groups', action='store', choices=['on', 'off'], help='Unique Groups.')

scim_delete_parser = argparse.ArgumentParser(prog='scim-delete', description='Delete SCIM endpoint.')
scim_delete_parser.add_argument('target', help='SCIM ID')
scim_delete_parser.add_argument('--force', '-f', dest='force', action='store_true', help='Delete with no confirmation')


def register_commands(commands):
    commands['scim'] = ScimCommand()


def register_command_info(aliases, command_info):
    command_info['scim'] = 'Manage SCIM endpoints'


class ScimCommand(GroupCommand):
    def __init__(self):
        super(ScimCommand, self).__init__()
        self.register_command('list', ScimListCommand())
        self.register_command('view', ScimViewCommand())
        self.register_command('create', ScimCreateCommand())
        self.register_command('edit', ScimEditCommand())
        self.register_command('delete', ScimDeleteCommand())
        self.default_verb = 'list'


class ScimListCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_list_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        table = []
        headers = ['SCIM ID', 'Node Name', 'Node ID', 'Prefix', 'Status', 'Last Synced']
        if params.enterprise and 'scims' in params.enterprise:
            for info in params.enterprise['scims']:
                node_id = info['node_id']
                last_synced = info.get('last_synced')
                if isinstance(last_synced, int):
                    dt = datetime.datetime.fromtimestamp(last_synced / 1000)
                    last_synced = dt.strftime('%c')
                else:
                    last_synced = str(last_synced or '')
                row = [info['scim_id'], self.get_node_path(params, node_id), node_id,
                       info.get('role_prefix') or '', info['status'], last_synced]
                table.append(row)
        dump_report_data(table, headers=headers)


class ScimCreateCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_create_parser

    def execute(self, params, node=None, **kwargs):
        if not node:
            logging.warning('\"--node\" option is required for \"create\" command')
            return
        nodes = list(self.resolve_nodes(params, node))
        if len(nodes) > 1:
            logging.warning('Node name \'%s\' is not unique. Use Node ID.', node)
            return
        elif len(nodes) == 0:
            logging.warning('Node name \'%s\' is not found', node)
            return

        matched_node = nodes[0]
        if not matched_node.get('parent_id'):
            logging.warning('Root node cannot be used for SCIM endpoint')
            return
        token = base64.urlsafe_b64encode(os.urandom(32)).decode().rstrip('=')
        rq = {
            'command': 'scim_add',
            'scim_id': self.get_enterprise_id(params),
            'node_id': matched_node['node_id'],
            'token': token,
        }
        prefix = kwargs.get('prefix')
        if prefix:
            rq['prefix'] = prefix
        if kwargs.get('unique_groups', '') == 'on':
            rq['unique_groups'] = True

        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('')
        logging.info('SCIM ID: %d', rq['scim_id'])
        logging.info('SCIM URL: %s', get_scim_url(params, matched_node['node_id']))
        logging.info('Provisioning Token: %s', token)
        logging.info('')
        return token


def find_scim(param, name):   # type: (KeeperParams, any) -> dict
    if not name:
        raise Exception('SCIM ID cannot be empty')
    if 'scims' in param.enterprise:
        try:
            scim_id = int(name)
            for scim in param.enterprise['scims']:
                if scim['scim_id'] == scim_id:
                    return scim
                if scim['node_id'] == scim_id:
                    return scim
        except:
            raise Exception('SCIM ID should be an integer: {0}'.format(name))
    raise Exception('SCIM endpoint with ID \"{0}\" not found'.format(name))


class ScimViewCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_view_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)

        logging.info('{0:>20s}: {1}'.format('SCIM ID', scim['scim_id']))
        node_id = scim['node_id']
        logging.info('{0:>20s}: {1}'.format('SCIM URL', get_scim_url(params, node_id)))
        logging.info('{0:>20s}: {1}'.format('Node ID', node_id))
        logging.info('{0:>20s}: {1}'.format('Node Name', self.get_node_path(params, node_id)))
        logging.info('{0:>20s}: {1}'.format('Status', scim['status']))
        logging.info('{0:>20s}: {1}'.format('Prefix', scim.get('role_prefix') or ''))
        logging.info('{0:>20s}: {1}'.format('Unique Groups', scim.get('unique_groups', False)))
        last_synced = scim.get('last_synced')
        if isinstance(last_synced, int):
            dt = datetime.datetime.fromtimestamp(last_synced / 1000)
            logging.info('{0:>20s}: {1}'.format('Last Synced', dt.strftime('%c')))


class ScimEditCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_edit_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        token = base64.urlsafe_b64encode(os.urandom(32)).decode().rstrip('=')
        rq = {
            'command': 'scim_update',
            'scim_id': scim['scim_id'],
            'token': token,
        }
        prefix = kwargs.get('prefix')
        if prefix:
            rq['prefix'] = prefix
        unique_groups = kwargs.get('unique_groups')
        if unique_groups:
            rq['unique_groups'] = unique_groups == 'on'

        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('')
        logging.info('SCIM ID: %d', scim['scim_id'])
        logging.info('SCIM URL: %s', get_scim_url(params, scim['node_id']))
        logging.info('Provisioning Token: %s', token)
        logging.info('')
        return token


class ScimDeleteCommand(EnterpriseCommand):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return scim_delete_parser

    def execute(self, params, target=None, **kwargs):
        scim = find_scim(params, target)
        answer = 'y' if kwargs.get('force') else \
            user_choice(bcolors.FAIL + bcolors.BOLD + '\nALERT!\n' + bcolors.ENDC +
                        'You are about to delete SCIM endpoint {0}'.format(scim['scim_id']) +
                        '\n\nDo you want to proceed with deletion?', 'yn', 'n')
        if answer.lower() != 'y':
            return

        rq = {
            'command': 'scim_delete',
            'scim_id': scim['scim_id'],
        }
        api.communicate(params, rq)
        api.query_enterprise(params)
        logging.info('SCIM endpoint \"%d\" at node \"%d\" deleted', scim['scim_id'], scim['node_id'])


def get_scim_url(params, node_id):  # type:  (KeeperParams, int) -> any
    p = urlparse(params.rest_context.server_base)
    return urlunparse((p.scheme, p.netloc, '/api/rest/scim/v2/' + str(node_id), None, None, None))



import json
import os
import io

from unittest import TestCase, mock

from data_vault import get_synced_params, VaultEnvironment
from helper import KeeperApiHelper

from keepercommander import api, utils, crypto, attachment, vault
from keepercommander.commands import recordv2, record
from keepercommander.error import CommandError


class TestRecord(TestCase):
    vault_env = VaultEnvironment()

    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = KeeperApiHelper.communicate_command

    def tearDown(self):
        mock.patch.stopall()

    def test_create_record(self):
        params = get_synced_params()
        record = vault.KeeperRecord.create(params, 'login')
        

    def test_add_command(self):
        params = get_synced_params()
        cmd = recordv2.RecordAddCommand()

        with mock.patch('keepercommander.api.sync_down'):
            KeeperApiHelper.communicate_expect(['record_add'])
            cmd.execute(params, force=True, title='New Record')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            KeeperApiHelper.communicate_expect(['record_add'])
            cmd.execute(params, login='login', password='password', url='url', custom='name1: value 1, name2: value 2', title='New Record')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            KeeperApiHelper.communicate_expect(['record_add'])
            cmd.execute(params, login='login', password='password', url='url', custom=[{'name1': 'value 1', 'name2': 'value 2'}], title='New Record')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            with mock.patch('builtins.input', return_value='Input Data'):
                KeeperApiHelper.communicate_expect(['record_add'])
                cmd.execute(params, force=True, title='New Record')
                self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_add_command_check_request(self):
        params = get_synced_params()
        cmd = recordv2.RecordAddCommand()

        def check_record(rq):
            rq['command'] = 'record_add'
            record_key = api.decrypt_data(rq['record_key'], self.vault_env.data_key)
            data_bytes = api.decrypt_data(rq['data'], record_key)
            data = json.loads(data_bytes.decode('utf-8'))
            self.assertEqual(data['title'], 'data')
            self.assertEqual(data['title'], data['link'])
            self.assertEqual(data['secret1'], data['secret2'])
            self.assertEqual(len(data['custom']), 2)

        with mock.patch('builtins.input', return_value='data'), mock.patch('keepercommander.api.sync_down'):
            KeeperApiHelper.communicate_expect([check_record])
            cmd.execute(params, custom='name1: value 1, name2: value 2')
            self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_remove_command_from_root(self):
        params = get_synced_params()
        cmd = recordv2.RecordRemoveCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        rec = api.get_record(params, record_uid)

        def pre_delete_command(rq):
            self.assertEqual(rq['command'], 'pre_delete')
            return {
                'pre_delete_response': {
                    'would_delete': {
                        'deletion_summary': ['delete all']
                    },

                    'pre_delete_token': 'token'
                }
            }

        with mock.patch('keepercommander.commands.base.user_choice') as choice_mock:
            choice_mock.return_value = KeyboardInterrupt()
            KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
            cmd.execute(params, force=True, record=rec.record_uid)
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
            cmd.execute(params, force=True, record=rec.title)
            self.assertTrue(KeeperApiHelper.is_expect_empty())

            with mock.patch('builtins.print'):
                choice_mock.return_value = 'y'
                KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
                cmd.execute(params, record=rec.record_uid)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                KeeperApiHelper.communicate_expect([pre_delete_command, 'delete'])
                cmd.execute(params, record=rec.title)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                choice_mock.return_value = 'n'
                KeeperApiHelper.communicate_expect([pre_delete_command])
                cmd.execute(params, record=rec.record_uid)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

                KeeperApiHelper.communicate_expect([pre_delete_command])
                cmd.execute(params, record=rec.title)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_search_command(self):
        params = get_synced_params()
        cmd = record.SearchCommand()

        with mock.patch('builtins.print'):
            cmd.execute(params, pattern='.*')
            cmd.execute(params, pattern='Non-existing-name')

    def test_record_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListCommand()

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, pattern='record')
            mock_print.assert_called()

            mock_print.reset_mock()
            cmd.execute(params, pattern='INVALID')
            mock_print.assert_not_called()

    def test_shared_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListSfCommand()

        with mock.patch('builtins.print') as mock_print:
            cmd.execute(params, pattern='folder')
            mock_print.assert_called()

            mock_print.reset_mock()
            cmd.execute(params, pattern='INVALID')
            mock_print.assert_not_called()

    def test_team_list_command(self):
        params = get_synced_params()
        cmd = record.RecordListTeamCommand()

        def get_available_teams(rq):
            self.assertEqual(rq['command'], 'get_available_teams')
            return {
                'teams': [
                    {
                        'team_uid': api.generate_record_uid(),
                        'team_name': 'Team 1'
                    },
                    {
                        'team_uid': api.generate_record_uid(),
                        'team_name': 'Team 2'
                    }
                ]
            }
        KeeperApiHelper.communicate_expect([get_available_teams])
        with mock.patch('builtins.print'):
            cmd.execute(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_get_record_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        with mock.patch('builtins.print'), mock.patch('keepercommander.api.get_record_shares'):
            cmd.execute(params, uid=record_uid)
            cmd.execute(params, format='json', uid=record_uid)

    def test_get_shared_folder_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        shared_folder_uid = next(iter(params.shared_folder_cache))
        with mock.patch('builtins.print'):
            cmd.execute(params, uid=shared_folder_uid)
            cmd.execute(params, format='json', uid=shared_folder_uid)

    def test_get_team_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        team_uid = next(iter(params.team_cache))
        with mock.patch('builtins.print'):
            cmd.execute(params, uid=team_uid)
            cmd.execute(params, format='json', uid=team_uid)

    def test_get_invalid_uid(self):
        params = get_synced_params()
        cmd = record.RecordGetUidCommand()

        with self.assertRaises(CommandError):
            cmd.execute(params, uid='invalid')

    def test_append_notes_command(self):
        params = get_synced_params()
        cmd = recordv2.RecordAppendNotesCommand()

        record_uid = next(iter(params.subfolder_record_cache['']))
        with mock.patch('keepercommander.api.update_record'):
            cmd.execute(params, notes='notes', record=record_uid)

            with mock.patch('builtins.input', return_value='data'):
                cmd.execute(params, record=record_uid)

        with self.assertRaises(CommandError):
            cmd.execute(params, notes='notes', record='invalid')

    def test_download_attachment_command(self):
        params = get_synced_params()
        cmd = recordv2.RecordDownloadAttachmentCommand()

        records = [x for x in params.record_cache.values() if len(x['extra_unencrypted']) > 10]
        rec = records[0]
        record_uid = rec['record_uid']
        extra = json.loads(rec['extra_unencrypted'].decode('utf-8'))
        attachments = {}  # type: dict[any, tuple[attachment.AttachmentDownloadRequest, bytes]]
        for file in extra['files']:
            atta_id = file['id']   # type: str
            key = utils.base64_url_decode(file['key'])
            body_encoded = crypto.encrypt_aes_v1(os.urandom(file['size']), key)
            rq = attachment.AttachmentDownloadRequest()
            rq.title = 'Title'
            rq.url = f'https://keepersecurity.com/files/{atta_id}'
            rq.is_gcm_encrypted = False
            rq.encryption_key = key

            attachments[atta_id] = (rq, body_encoded)

        def prepare_download(params, record_uid):
            return (x[0] for x in attachments.values())

        def requests_get(url, **kwargs):
            body = next((x[1] for x in attachments.values() if x[0].url == url), None)
            if not body:
                raise Exception(f'URL \"{url}\" not found.')
            rs = mock.Mock()
            rs.status_code = 200
            stream = io.BytesIO(body)
            rs.raw = stream
            rs.__enter__ = mock.Mock(return_value=rs)
            rs.__exit__ = mock.Mock(return_value=None)
            return rs

        with mock.patch('keepercommander.attachment.prepare_attachment_download', side_effect=prepare_download), \
                mock.patch('requests.get', side_effect=requests_get), \
                mock.patch('builtins.open', mock.mock_open()), \
                mock.patch('os.path.abspath', return_value='/file_name'):
            cmd.execute(params, record=record_uid)

    def test_upload_attachment_command(self):
        params = get_synced_params()
        cmd = recordv2.RecordUploadAttachmentCommand()

        record_uid = next(iter([x['record_uid'] for x in params.record_cache.values() if len(x['extra_unencrypted']) > 10]))
        rec = api.get_record(params, record_uid)

        with self.assertRaises(CommandError):
            cmd.execute(params, record=rec.title)

        def request_upload(rq):
            self.assertEqual(rq['command'], 'request_upload')
            return {
                'file_uploads': [{
                    'max_size': 1000000,
                    'url': 'https://keepersecurity.com/uploads/',
                    'success_status_code': 201,
                    'file_id': 'ABCDEF%.2d' % x,
                    'file_parameter': 'file',
                    'parameters': {'a':'b'}
                } for x in range((rq.get('file_count') or 0) + (rq.get('thumbnail_count') or 0))]
            }

        def request_http_post(url, **kwargs):
            attachment = mock.Mock()
            attachment.status_code = 201
            return attachment

        with mock.patch('requests.post', side_effect=request_http_post), \
                mock.patch('builtins.open', mock.mock_open(read_data=b'data')) as m_open, \
                mock.patch('os.path.isfile', return_value=True), \
                mock.patch('os.path.getsize') as mock_getsize:

            m_open.return_value.tell = lambda: 4
            mock_getsize.return_value = 1000000000
            with self.assertRaises(CommandError):
                KeeperApiHelper.communicate_expect([request_upload])
                cmd.execute(params, file=['file.data'], record=record_uid)
                self.assertTrue(KeeperApiHelper.is_expect_empty())

            KeeperApiHelper.communicate_expect([request_upload, 'record_update'])
            m_open.return_value.tell = lambda: 4
            mock_getsize.return_value = 1000
            cmd.execute(params, file=['file.data'], record=record_uid)
            self.assertTrue(KeeperApiHelper.is_expect_empty())

    def test_delete_attachment_command(self):
        params = get_synced_params()
        cmd = recordv2.RecordDeleteAttachmentCommand()

        record_uid = next(iter([x['record_uid'] for x in params.record_cache.values() if len(x['extra_unencrypted']) > 10]))
        rec = api.get_record(params, record_uid)

        KeeperApiHelper.communicate_expect(['record_update'])
        cmd.execute(params, name=[rec.attachments[0]['id']], record=rec.title)
        self.assertTrue(KeeperApiHelper.is_expect_empty())



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import base64
import logging
from urllib.parse import urlparse, urlunparse
from typing import Iterator, Tuple, Optional, List, Callable, Union

from .constants import KEEPER_PUBLIC_HOSTS
from . import api, crypto, utils, rest_api
from .proto import breachwatch_pb2 as breachwatch_proto, client_pb2 as client_proto
from .proto import APIRequest_pb2 as api_request_proto
from .error import KeeperApiError
from .record import Record
from .params import KeeperParams


class BreachWatch(object):
    def __init__(self):
        self.rest_api = None
        self.domain_token = None
        self.email_token = None
        self.password_token = None
        self.send_audit_events = False

    def scan_password(self, params, password, euid=None):
        # type: (KeeperParams, str, Optional[bytes]) -> Optional[breachwatch_proto.HashStatus]

        bw_hash = utils.breach_watch_hash(password)
        if not euid:
            score = utils.password_score(password)
            if score < 40:
                result = breachwatch_proto.HashStatus()
                result.hash1 = bw_hash
                result.breachDetected = True
                return result

        self._ensure_init(params)
        check = breachwatch_proto.HashCheck()
        check.hash1 = bw_hash
        if euid:
            check.euid = euid
        rq = breachwatch_proto.BreachWatchStatusRequest()
        rq.hashCheck.append(check)
        rs = self._execute_status(rq)
        return rs.hashStatus[0]

    def scan_passwords(self, params, passwords):  # type: (any, Iterator[str]) -> Iterator[Tuple[str, any]]
        results = {}
        hashes = {}
        if passwords:
            for password in passwords:
                score = utils.password_score(password)
                bw_hash = utils.breach_watch_hash(password)
                if score >= 40:
                    hashes[bw_hash] = password
                else:
                    status = breachwatch_proto.HashStatus()
                    status.hash1 = bw_hash
                    status.breachDetected = True
                    results[password] = status
        if len(hashes) > 0:
            rq = breachwatch_proto.BreachWatchStatusRequest()
            for hash in hashes:
                check = breachwatch_proto.HashCheck()
                check.hash1 = hash
                rq.hashCheck.append(check)
            self._ensure_init(params)
            rs = self._execute_status(rq)
            for status in rs.hashStatus:
                results[hashes[status.hash1]] = status

        for password in results:
            yield password, results[password]

    def scan_and_store_record_status(self, params, record_uid):  # type: (KeeperParams, str) -> None
        record = api.get_record(params, record_uid)
        if not record:
            return

        bw_record = params.breach_watch_records.get(record_uid) if params.breach_watch_records else None
        if record.password:
            euid = None
            if bw_record:
                data_obj = bw_record.get('data_unencrypted')
                if data_obj and 'passwords' in data_obj:
                    password = next((x for x in data_obj['passwords'] if x.get('value', '') == record.password), None)
                    if password:
                        return
                    euid = next((base64.b64decode(x['euid']) for x in data_obj['passwords'] if 'euid' in x), None)

            hash_status = self.scan_password(params, record.password, euid)
            if hash_status.breachDetected:
                logging.info('High-Risk password detected')
                if self.send_audit_events:
                    params.queue_audit_event('bw_record_high_risk')

            bwrq = breachwatch_proto.BreachWatchRecordRequest()
            bwrq.recordUid = utils.base64_url_decode(record_uid)
            bwrq.breachWatchInfoType = breachwatch_proto.RECORD
            bwrq.updateUserWhoScanned = True
            bw_password = client_proto.BWPassword()
            bw_password.value = record.password
            bw_password.status = client_proto.WEAK if hash_status.breachDetected else client_proto.GOOD
            bw_password.euid = hash_status.euid
            bw_data = client_proto.BreachWatchData()
            bw_data.passwords.append(bw_password)
            data = bw_data.SerializeToString()
            try:
                record_key = params.record_cache[record_uid]['record_key_unencrypted']
                bwrq.encryptedData = crypto.encrypt_aes_v2(data, record_key)
                rq = breachwatch_proto.BreachWatchUpdateRequest()
                rq.breachWatchRecordRequest.append(bwrq)
                rs = api.communicate_rest(params, rq, 'breachwatch/update_record_data',
                                          rs_type=breachwatch_proto.BreachWatchUpdateResponse)
                status = rs.breachWatchRecordStatus[0]
                if status.reason:
                    raise Exception(status.reason)
            except Exception as e:
                logging.warning('BreachWatch: %s', str(e))

    def delete_euids(self, params, euids):
        self._ensure_init(params)
        while euids:
            chunk = euids[:999]
            euids = euids[999:]
            rq = breachwatch_proto.BreachWatchStatusRequest()
            rq.removedEuid.extend(chunk)
            self._execute_status(rq)

    def _execute_status(self, rq):
        rq.anonymizedToken = self.password_token
        api_request_payload = api_request_proto.ApiRequestPayload()
        api_request_payload.payload = rq.SerializeToString()
        rs = rest_api.execute_rest(self.rest_api, 'breachwatch/status', api_request_payload)
        if isinstance(rs, bytes):
            bw_rs = breachwatch_proto.BreachWatchStatusResponse()
            bw_rs.ParseFromString(rs)
            return bw_rs
        elif isinstance(rs, dict):
            raise KeeperApiError(rs['error'], rs['message'])
        raise KeeperApiError('Error', 'breachwatch/status')

    def _ensure_init(self, params):
        if self.rest_api:
            if not self.password_token:
                raise KeeperApiError('not_initialized', 'BreachWatch init error.')
            return

        url_comp = urlparse(params.rest_context.server_base)
        us_server = KEEPER_PUBLIC_HOSTS['US']
        if not url_comp.netloc.endswith(us_server):
            for region in KEEPER_PUBLIC_HOSTS.values():
                if url_comp.netloc.endswith(region):
                    url_comp = url_comp._replace(netloc=url_comp.netloc[0: -len(region)] + us_server)
                    break
        bw_endpoint = rest_api.RestApiContext(server=urlunparse(url_comp), locale=params.rest_context.locale)
        bw_endpoint.server_key_id = params.rest_context.server_key_id
        self.rest_api = bw_endpoint

        rs = api.communicate_rest(params, None, 'breachwatch/initialize',
                                  rs_type=breachwatch_proto.BreachWatchTokenResponse)
        if rs.clientEncrypted:
            enc_token = rs.breachWatchToken
            breach_watch_token = crypto.decrypt_aes_v2(enc_token, params.data_key)
        else:
            breach_watch_token = rs.breachWatchToken
            enc_token = crypto.encrypt_aes_v2(breach_watch_token, params.data_key)
            rq = breachwatch_proto.BreachWatchTokenRequest()
            rq.breachWatchToken = enc_token
            api.communicate_rest(params, rq, 'breachwatch/save_token')

        rq = breachwatch_proto.BreachWatchTokenRequest()
        rq.breachWatchToken = breach_watch_token
        rs = api.communicate_rest(params, rq, 'breachwatch/anonymize_token',
                                  rs_type=breachwatch_proto.AnonymizedTokenResponse)
        self.domain_token = rs.domainToken
        self.email_token = rs.emailToken
        self.password_token = rs.passwordToken

        if params.enforcements:
            pass

    @staticmethod
    def get_record_status(params, record_uid):  # type: (KeeperParams, str) -> Optional[dict]
        if not params.breach_watch_records:
            return
        bw_record = params.breach_watch_records.get(record_uid)
        if bw_record:
            data_obj = bw_record['data_unencrypted']
            if data_obj and 'passwords' in data_obj:
                record = api.get_record(params, record_uid)
                return next((x for x in data_obj['passwords'] if x.get('value', '') == record.password), None)

    @staticmethod
    def get_records(params, callback, owned=False):
        # type: (KeeperParams, Callable[[Record, Optional[dict]], bool], bool) -> Iterator[Tuple[Record, Optional[dict]]]
        if not params.record_cache:
            return
        for record_uid in params.record_cache:
            record = api.get_record(params, record_uid)
            if not record:
                continue
            if not record.password:
                continue
            if owned:
                if record_uid not in params.meta_data_cache:
                    continue
                meta_data = params.meta_data_cache[record_uid]
                if not meta_data.get('owner'):
                    continue
            password_dict = None
            if params.breach_watch_records:
                bwr = params.breach_watch_records.get(record_uid)
                data_obj = bwr['data_unencrypted'] if bwr else None
                if data_obj and 'passwords' in data_obj:
                    password_dict = next((x for x in data_obj['passwords'] if x.get('value', '') == record.password), None)
            if callback(record, password_dict):
                yield record, password_dict

    @staticmethod
    def get_records_to_scan(params):  # type: (KeeperParams) -> Iterator[Record]
        yield from BreachWatch.get_records(params, lambda r, b: b is None, owned=True)

    @staticmethod
    def check_status(bwr, statuses):
        if isinstance(bwr, dict) and isinstance(statuses, set):
            return bwr.get('status', '').casefold() in statuses
        return False

    @staticmethod
    def get_records_by_status(params, status, owned=False):
        # type: (KeeperParams, Optional[str, List[str]], bool) -> Iterator[Record]
        statuses = set()
        if status:
            if isinstance(status, list):
                statuses.update((x.casefold() for x in status if isinstance(x, str)))
            elif isinstance(status, str):
                statuses.add(status.casefold())
        else:
            statuses.update((x.casefold() for x in client_proto.BWStatus.keys()))

        yield from params.breach_watch.get_records(params, lambda r, b: BreachWatch.check_status(b, statuses), owned)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pam.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import enterprise_pb2 as enterprise__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\tpam.proto\x12\x06Router\x1a\x10\x65nterprise.proto\"\x83\x01\n\x13PAMRotationSchedule\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x18\n\x10\x63onfigurationUid\x18\x02 \x01(\x0c\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x14\n\x0cscheduleData\x18\x04 \x01(\t\x12\x12\n\nnoSchedule\x18\x05 \x01(\x08\"N\n\x1cPAMRotationSchedulesResponse\x12.\n\tschedules\x18\x01 \x03(\x0b\x32\x1b.Router.PAMRotationSchedule\"<\n\x13PAMOnlineController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x0e\n\x06\x63ookie\x18\x02 \x01(\t\"H\n\x14PAMOnlineControllers\x12\x30\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x1b.Router.PAMOnlineController\"9\n\x10PAMRotateRequest\x12\x12\n\nrequestUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\"D\n\x16PAMControllersResponse\x12*\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x15.Router.PAMController\"=\n\x13PAMRemoveController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\t\"O\n\x1bPAMRemoveControllerResponse\x12\x30\n\x0b\x63ontrollers\x18\x01 \x03(\x0b\x32\x1b.Router.PAMRemoveController\"@\n\x10PAMModifyRequest\x12,\n\noperations\x18\x01 \x03(\x0b\x32\x18.Router.PAMDataOperation\"\xa1\x01\n\x10PAMDataOperation\x12/\n\roperationType\x18\x01 \x01(\x0e\x32\x18.Router.PAMOperationType\x12\x33\n\rconfiguration\x18\x02 \x01(\x0b\x32\x1c.Router.PAMConfigurationData\x12\'\n\x07\x65lement\x18\x03 \x01(\x0b\x32\x16.Router.PAMElementData\"e\n\x14PAMConfigurationData\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\"E\n\x0ePAMElementData\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12\x11\n\tparentUid\x18\x02 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\x0c\"p\n\x19PAMElementOperationResult\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12.\n\x06result\x18\x02 \x01(\x0e\x32\x1e.Router.PAMOperationResultType\x12\x0f\n\x07message\x18\x03 \x01(\t\"E\n\x0fPAMModifyResult\x12\x32\n\x07results\x18\x01 \x03(\x0b\x32!.Router.PAMElementOperationResult\"{\n\nPAMElement\x12\x12\n\nelementUid\x18\x01 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x03 \x01(\x03\x12\x14\n\x0clastModified\x18\x04 \x01(\x03\x12$\n\x08\x63hildren\x18\x05 \x03(\x0b\x32\x12.Router.PAMElement\"#\n\x14PAMGenericUidRequest\x12\x0b\n\x03uid\x18\x01 \x01(\x0c\"%\n\x15PAMGenericUidsRequest\x12\x0c\n\x04uids\x18\x01 \x03(\x0c\"\xae\x01\n\x10PAMConfiguration\x12\x18\n\x10\x63onfigurationUid\x18\x01 \x01(\x0c\x12\x0e\n\x06nodeId\x18\x02 \x01(\x03\x12\x15\n\rcontrollerUid\x18\x03 \x01(\x0c\x12\x0c\n\x04\x64\x61ta\x18\x04 \x01(\x0c\x12\x0f\n\x07\x63reated\x18\x05 \x01(\x03\x12\x14\n\x0clastModified\x18\x06 \x01(\x03\x12$\n\x08\x63hildren\x18\x07 \x03(\x0b\x32\x12.Router.PAMElement\"E\n\x11PAMConfigurations\x12\x30\n\x0e\x63onfigurations\x18\x01 \x03(\x0b\x32\x18.Router.PAMConfiguration\"\xe8\x01\n\rPAMController\x12\x15\n\rcontrollerUid\x18\x01 \x01(\x0c\x12\x16\n\x0e\x63ontrollerName\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65viceToken\x18\x03 \x01(\t\x12\x12\n\ndeviceName\x18\x04 \x01(\t\x12\x0e\n\x06nodeId\x18\x05 \x01(\x03\x12\x0f\n\x07\x63reated\x18\x06 \x01(\x03\x12\x14\n\x0clastModified\x18\x07 \x01(\x03\x12\x16\n\x0e\x61pplicationUid\x18\x08 \x01(\x0c\x12\x30\n\rappClientType\x18\t \x01(\x0e\x32\x19.Enterprise.AppClientType\"%\n\x12\x43ontrollerResponse\x12\x0f\n\x07payload\x18\x01 \x01(\t*@\n\x10PAMOperationType\x12\x07\n\x03\x41\x44\x44\x10\x00\x12\n\n\x06UPDATE\x10\x01\x12\x0b\n\x07REPLACE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03*p\n\x16PAMOperationResultType\x12\x0f\n\x0bPOT_SUCCESS\x10\x00\x12\x15\n\x11POT_UNKNOWN_ERROR\x10\x01\x12\x16\n\x12POT_ALREADY_EXISTS\x10\x02\x12\x16\n\x12POT_DOES_NOT_EXIST\x10\x03*H\n\x15\x43ontrollerMessageType\x12\x0f\n\x0b\x43MT_GENERAL\x10\x00\x12\x0e\n\nCMT_ROTATE\x10\x01\x12\x0e\n\nCMT_STREAM\x10\x02\x42\x1f\n\x18\x63om.keepersecurity.protoB\x03PAMb\x06proto3')

_PAMOPERATIONTYPE = DESCRIPTOR.enum_types_by_name['PAMOperationType']
PAMOperationType = enum_type_wrapper.EnumTypeWrapper(_PAMOPERATIONTYPE)
_PAMOPERATIONRESULTTYPE = DESCRIPTOR.enum_types_by_name['PAMOperationResultType']
PAMOperationResultType = enum_type_wrapper.EnumTypeWrapper(_PAMOPERATIONRESULTTYPE)
_CONTROLLERMESSAGETYPE = DESCRIPTOR.enum_types_by_name['ControllerMessageType']
ControllerMessageType = enum_type_wrapper.EnumTypeWrapper(_CONTROLLERMESSAGETYPE)
ADD = 0
UPDATE = 1
REPLACE = 2
DELETE = 3
POT_SUCCESS = 0
POT_UNKNOWN_ERROR = 1
POT_ALREADY_EXISTS = 2
POT_DOES_NOT_EXIST = 3
CMT_GENERAL = 0
CMT_ROTATE = 1
CMT_STREAM = 2


_PAMROTATIONSCHEDULE = DESCRIPTOR.message_types_by_name['PAMRotationSchedule']
_PAMROTATIONSCHEDULESRESPONSE = DESCRIPTOR.message_types_by_name['PAMRotationSchedulesResponse']
_PAMONLINECONTROLLER = DESCRIPTOR.message_types_by_name['PAMOnlineController']
_PAMONLINECONTROLLERS = DESCRIPTOR.message_types_by_name['PAMOnlineControllers']
_PAMROTATEREQUEST = DESCRIPTOR.message_types_by_name['PAMRotateRequest']
_PAMCONTROLLERSRESPONSE = DESCRIPTOR.message_types_by_name['PAMControllersResponse']
_PAMREMOVECONTROLLER = DESCRIPTOR.message_types_by_name['PAMRemoveController']
_PAMREMOVECONTROLLERRESPONSE = DESCRIPTOR.message_types_by_name['PAMRemoveControllerResponse']
_PAMMODIFYREQUEST = DESCRIPTOR.message_types_by_name['PAMModifyRequest']
_PAMDATAOPERATION = DESCRIPTOR.message_types_by_name['PAMDataOperation']
_PAMCONFIGURATIONDATA = DESCRIPTOR.message_types_by_name['PAMConfigurationData']
_PAMELEMENTDATA = DESCRIPTOR.message_types_by_name['PAMElementData']
_PAMELEMENTOPERATIONRESULT = DESCRIPTOR.message_types_by_name['PAMElementOperationResult']
_PAMMODIFYRESULT = DESCRIPTOR.message_types_by_name['PAMModifyResult']
_PAMELEMENT = DESCRIPTOR.message_types_by_name['PAMElement']
_PAMGENERICUIDREQUEST = DESCRIPTOR.message_types_by_name['PAMGenericUidRequest']
_PAMGENERICUIDSREQUEST = DESCRIPTOR.message_types_by_name['PAMGenericUidsRequest']
_PAMCONFIGURATION = DESCRIPTOR.message_types_by_name['PAMConfiguration']
_PAMCONFIGURATIONS = DESCRIPTOR.message_types_by_name['PAMConfigurations']
_PAMCONTROLLER = DESCRIPTOR.message_types_by_name['PAMController']
_CONTROLLERRESPONSE = DESCRIPTOR.message_types_by_name['ControllerResponse']
PAMRotationSchedule = _reflection.GeneratedProtocolMessageType('PAMRotationSchedule', (_message.Message,), {
  'DESCRIPTOR' : _PAMROTATIONSCHEDULE,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMRotationSchedule)
  })
_sym_db.RegisterMessage(PAMRotationSchedule)

PAMRotationSchedulesResponse = _reflection.GeneratedProtocolMessageType('PAMRotationSchedulesResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAMROTATIONSCHEDULESRESPONSE,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMRotationSchedulesResponse)
  })
_sym_db.RegisterMessage(PAMRotationSchedulesResponse)

PAMOnlineController = _reflection.GeneratedProtocolMessageType('PAMOnlineController', (_message.Message,), {
  'DESCRIPTOR' : _PAMONLINECONTROLLER,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMOnlineController)
  })
_sym_db.RegisterMessage(PAMOnlineController)

PAMOnlineControllers = _reflection.GeneratedProtocolMessageType('PAMOnlineControllers', (_message.Message,), {
  'DESCRIPTOR' : _PAMONLINECONTROLLERS,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMOnlineControllers)
  })
_sym_db.RegisterMessage(PAMOnlineControllers)

PAMRotateRequest = _reflection.GeneratedProtocolMessageType('PAMRotateRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAMROTATEREQUEST,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMRotateRequest)
  })
_sym_db.RegisterMessage(PAMRotateRequest)

PAMControllersResponse = _reflection.GeneratedProtocolMessageType('PAMControllersResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAMCONTROLLERSRESPONSE,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMControllersResponse)
  })
_sym_db.RegisterMessage(PAMControllersResponse)

PAMRemoveController = _reflection.GeneratedProtocolMessageType('PAMRemoveController', (_message.Message,), {
  'DESCRIPTOR' : _PAMREMOVECONTROLLER,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMRemoveController)
  })
_sym_db.RegisterMessage(PAMRemoveController)

PAMRemoveControllerResponse = _reflection.GeneratedProtocolMessageType('PAMRemoveControllerResponse', (_message.Message,), {
  'DESCRIPTOR' : _PAMREMOVECONTROLLERRESPONSE,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMRemoveControllerResponse)
  })
_sym_db.RegisterMessage(PAMRemoveControllerResponse)

PAMModifyRequest = _reflection.GeneratedProtocolMessageType('PAMModifyRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAMMODIFYREQUEST,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMModifyRequest)
  })
_sym_db.RegisterMessage(PAMModifyRequest)

PAMDataOperation = _reflection.GeneratedProtocolMessageType('PAMDataOperation', (_message.Message,), {
  'DESCRIPTOR' : _PAMDATAOPERATION,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMDataOperation)
  })
_sym_db.RegisterMessage(PAMDataOperation)

PAMConfigurationData = _reflection.GeneratedProtocolMessageType('PAMConfigurationData', (_message.Message,), {
  'DESCRIPTOR' : _PAMCONFIGURATIONDATA,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMConfigurationData)
  })
_sym_db.RegisterMessage(PAMConfigurationData)

PAMElementData = _reflection.GeneratedProtocolMessageType('PAMElementData', (_message.Message,), {
  'DESCRIPTOR' : _PAMELEMENTDATA,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMElementData)
  })
_sym_db.RegisterMessage(PAMElementData)

PAMElementOperationResult = _reflection.GeneratedProtocolMessageType('PAMElementOperationResult', (_message.Message,), {
  'DESCRIPTOR' : _PAMELEMENTOPERATIONRESULT,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMElementOperationResult)
  })
_sym_db.RegisterMessage(PAMElementOperationResult)

PAMModifyResult = _reflection.GeneratedProtocolMessageType('PAMModifyResult', (_message.Message,), {
  'DESCRIPTOR' : _PAMMODIFYRESULT,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMModifyResult)
  })
_sym_db.RegisterMessage(PAMModifyResult)

PAMElement = _reflection.GeneratedProtocolMessageType('PAMElement', (_message.Message,), {
  'DESCRIPTOR' : _PAMELEMENT,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMElement)
  })
_sym_db.RegisterMessage(PAMElement)

PAMGenericUidRequest = _reflection.GeneratedProtocolMessageType('PAMGenericUidRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAMGENERICUIDREQUEST,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMGenericUidRequest)
  })
_sym_db.RegisterMessage(PAMGenericUidRequest)

PAMGenericUidsRequest = _reflection.GeneratedProtocolMessageType('PAMGenericUidsRequest', (_message.Message,), {
  'DESCRIPTOR' : _PAMGENERICUIDSREQUEST,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMGenericUidsRequest)
  })
_sym_db.RegisterMessage(PAMGenericUidsRequest)

PAMConfiguration = _reflection.GeneratedProtocolMessageType('PAMConfiguration', (_message.Message,), {
  'DESCRIPTOR' : _PAMCONFIGURATION,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMConfiguration)
  })
_sym_db.RegisterMessage(PAMConfiguration)

PAMConfigurations = _reflection.GeneratedProtocolMessageType('PAMConfigurations', (_message.Message,), {
  'DESCRIPTOR' : _PAMCONFIGURATIONS,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMConfigurations)
  })
_sym_db.RegisterMessage(PAMConfigurations)

PAMController = _reflection.GeneratedProtocolMessageType('PAMController', (_message.Message,), {
  'DESCRIPTOR' : _PAMCONTROLLER,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.PAMController)
  })
_sym_db.RegisterMessage(PAMController)

ControllerResponse = _reflection.GeneratedProtocolMessageType('ControllerResponse', (_message.Message,), {
  'DESCRIPTOR' : _CONTROLLERRESPONSE,
  '__module__' : 'pam_pb2'
  # @@protoc_insertion_point(class_scope:Router.ControllerResponse)
  })
_sym_db.RegisterMessage(ControllerResponse)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\003PAM'
  _PAMOPERATIONTYPE._serialized_start=1974
  _PAMOPERATIONTYPE._serialized_end=2038
  _PAMOPERATIONRESULTTYPE._serialized_start=2040
  _PAMOPERATIONRESULTTYPE._serialized_end=2152
  _CONTROLLERMESSAGETYPE._serialized_start=2154
  _CONTROLLERMESSAGETYPE._serialized_end=2226
  _PAMROTATIONSCHEDULE._serialized_start=40
  _PAMROTATIONSCHEDULE._serialized_end=171
  _PAMROTATIONSCHEDULESRESPONSE._serialized_start=173
  _PAMROTATIONSCHEDULESRESPONSE._serialized_end=251
  _PAMONLINECONTROLLER._serialized_start=253
  _PAMONLINECONTROLLER._serialized_end=313
  _PAMONLINECONTROLLERS._serialized_start=315
  _PAMONLINECONTROLLERS._serialized_end=387
  _PAMROTATEREQUEST._serialized_start=389
  _PAMROTATEREQUEST._serialized_end=446
  _PAMCONTROLLERSRESPONSE._serialized_start=448
  _PAMCONTROLLERSRESPONSE._serialized_end=516
  _PAMREMOVECONTROLLER._serialized_start=518
  _PAMREMOVECONTROLLER._serialized_end=579
  _PAMREMOVECONTROLLERRESPONSE._serialized_start=581
  _PAMREMOVECONTROLLERRESPONSE._serialized_end=660
  _PAMMODIFYREQUEST._serialized_start=662
  _PAMMODIFYREQUEST._serialized_end=726
  _PAMDATAOPERATION._serialized_start=729
  _PAMDATAOPERATION._serialized_end=890
  _PAMCONFIGURATIONDATA._serialized_start=892
  _PAMCONFIGURATIONDATA._serialized_end=993
  _PAMELEMENTDATA._serialized_start=995
  _PAMELEMENTDATA._serialized_end=1064
  _PAMELEMENTOPERATIONRESULT._serialized_start=1066
  _PAMELEMENTOPERATIONRESULT._serialized_end=1178
  _PAMMODIFYRESULT._serialized_start=1180
  _PAMMODIFYRESULT._serialized_end=1249
  _PAMELEMENT._serialized_start=1251
  _PAMELEMENT._serialized_end=1374
  _PAMGENERICUIDREQUEST._serialized_start=1376
  _PAMGENERICUIDREQUEST._serialized_end=1411
  _PAMGENERICUIDSREQUEST._serialized_start=1413
  _PAMGENERICUIDSREQUEST._serialized_end=1450
  _PAMCONFIGURATION._serialized_start=1453
  _PAMCONFIGURATION._serialized_end=1627
  _PAMCONFIGURATIONS._serialized_start=1629
  _PAMCONFIGURATIONS._serialized_end=1698
  _PAMCONTROLLER._serialized_start=1701
  _PAMCONTROLLER._serialized_end=1933
  _CONTROLLERRESPONSE._serialized_start=1935
  _CONTROLLERRESPONSE._serialized_end=1972
# @@protoc_insertion_point(module_scope)



# __  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import csv
from urllib.parse import urlunsplit, urlencode

from ..importer import BaseFileImporter, Record, RecordField, FIELD_TYPE_ONE_TIME_CODE


# TOTP
TOTP_URL_SCHEME = 'otpauth'
TOTP_URL_NETLOC = 'totp'
TOTP_URL_PATH = '/lastpass_import'
TOTP_URL_QUERY_MAPPING = [
    ('algorithm', 'SHA1'),
    ('digits', '6'),
    ('period', '30')
]


'''
nickname       - title
username       - login
password       - password
url            - login_url
additionalInfo - notes
twofaSecret    - TOTP
status         - custom text field
'''


class MyKiCsvImporter(BaseFileImporter):

    def do_import(self, filename, **kwargs):
        with open(filename, 'r', encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                record = Record()
                record.title = row.get('nickname', '').strip()
                record.login = row.get('username', '').strip()
                record.password = row.get('password', '').strip()
                record.login_url = row.get('url', '').strip()
                record.notes = row.get('additionalInfo', '').strip()

                totp_secret = row.get('twofaSecret', '').strip().replace(' ', '')
                if len(totp_secret) > 0:
                    totp_query_string = urlencode([('secret', totp_secret)] + TOTP_URL_QUERY_MAPPING)
                    totp_url = urlunsplit((TOTP_URL_SCHEME, TOTP_URL_NETLOC, TOTP_URL_PATH, totp_query_string, ''))
                    record.fields.append(RecordField(type=FIELD_TYPE_ONE_TIME_CODE, value=totp_url))

                status = row.get('status', '').strip()
                if len(status) > 0:
                    record.fields.append(RecordField(type='text', label='status', value=status))

                yield record

    def extension(self):
        return 'csv'
